import typing
import random
import math

from ..api import op, weave_class
from .. import weave_types as types
import numpy as np
from .. import timestamp as weave_timestamp

binary_number_op_input_type = {
    "lhs": types.Number(),
    "rhs": types.optional(types.Number()),
}


def _set_add(lhs, rhs, v, action=None):
    return v - rhs


@weave_class(weave_type=types.Number)
class Number(object):
    @op(
        name="number-add",
        # Just to show that mutations can work in interesting ways.
        setter=_set_add,
        input_type=binary_number_op_input_type,
        output_type=types.Number(),
    )
    def __add__(lhs, rhs):
        if rhs == None or lhs == None:
            return None
        return lhs + rhs

    @op(
        name="number-sub",
        input_type=binary_number_op_input_type,
        output_type=types.Number(),
    )
    def __sub__(lhs, rhs):
        if rhs == None or lhs == None:
            return None
        return lhs - rhs

    @op(
        name="number-mult",
        input_type=binary_number_op_input_type,
        output_type=types.Number(),
    )
    def __mul__(lhs, rhs):
        if rhs == None or lhs == None:
            return None
        return lhs * rhs

    @op(
        name="number-div",
        input_type=binary_number_op_input_type,
        output_type=types.Number(),
    )
    def __truediv__(lhs, rhs):
        rhs = rhs or 0
        if rhs == 0:
            return math.inf
        return lhs / rhs

    @op(
        name="number-floorDiv",
        input_type=binary_number_op_input_type,
        output_type=types.Number(),
    )
    def __floordiv__(lhs, rhs):
        rhs = rhs or 0
        if rhs == 0:
            return math.inf
        return lhs // rhs

    @op(
        name="number-modulo",
        input_type=binary_number_op_input_type,
        output_type=types.Number(),
    )
    def __mod__(lhs, rhs):
        rhs = rhs or 0
        if rhs == 0:
            return 0
        return lhs % rhs

    @op(
        name="number-powBinary",
        input_type=binary_number_op_input_type,
        output_type=types.Number(),
    )
    def __pow__(lhs, rhs):
        rhs = rhs or 0
        return lhs**rhs

    @op(
        name="number-equal",
        input_type={
            **binary_number_op_input_type,
            "lhs": types.optional(types.Number()),
        },
        output_type=types.Boolean(),
    )
    def __eq__(lhs, rhs):
        return lhs == rhs

    @op(
        name="number-notEqual",
        input_type={
            **binary_number_op_input_type,
            "lhs": types.optional(types.Number()),
        },
        output_type=types.Boolean(),
    )
    def __ne__(lhs, rhs):
        return lhs != rhs

    @op(
        name="number-less",
        input_type=binary_number_op_input_type,
        output_type=types.Boolean(),
    )
    def __lt__(lhs, rhs):
        rhs = rhs or 0
        return lhs < rhs

    @op(
        name="number-greater",
        input_type=binary_number_op_input_type,
        output_type=types.Boolean(),
    )
    def __gt__(lhs, rhs):
        rhs = rhs or 0
        return lhs > rhs

    @op(
        name="number-lessEqual",
        input_type=binary_number_op_input_type,
        output_type=types.Boolean(),
    )
    def __le__(lhs, rhs):
        rhs = rhs or 0
        return lhs <= rhs

    @op(
        name="number-greaterEqual",
        input_type=binary_number_op_input_type,
        output_type=types.Boolean(),
    )
    def __ge__(lhs, rhs):
        rhs = rhs or 0
        return lhs >= rhs

    @op(
        name="number-pow",
        input_type=binary_number_op_input_type,
        output_type=types.Number(),
    )
    def pow(lhs, rhs):
        rhs = rhs or 0
        return lhs**rhs

    @op(
        name="number-negate",
        input_type={"val": types.Number()},
        output_type=types.Number(),
    )
    def __neg__(val):
        return val * -1

    @op(
        name="number-round",
        input_type={"val": types.Number()},
        output_type=types.Number(),
    )
    def __round__(val):
        return round(val)

    @op(
        name="number-floor",
        input_type={"number": types.Number()},
        output_type=types.Number(),
    )
    def floor(number):
        return math.floor(number)

    @op(
        name="number-ceil",
        input_type={"number": types.Number()},
        output_type=types.Number(),
    )
    def ceil(number):
        return math.ceil(number)

    @op(
        name="number-cos",
        input_type={"n": types.Number()},
        output_type=types.Number(),
    )
    def cos(n):
        return math.cos(n)

    @op(
        name="number-sin",
        input_type={"n": types.Number()},
        output_type=types.Number(),
    )
    def sin(n):
        return math.sin(n)

    @op(
        name="number-toTimestamp",
        input_type={"val": types.Number()},
        output_type=types.Timestamp(),
    )
    def to_timestamp(val):
        return weave_timestamp.ms_to_python_datetime(
            weave_timestamp.unitless_int_to_inferred_ms(val)
        )

    @op(
        name="number-toString",
        input_type={"val": types.Number()},
        output_type=types.String(),
    )
    def to_string(val):
        return str(val)

    @op(
        name="number-abs",
        input_type={"val": types.Number()},
        output_type=types.Number(),
    )
    def abs(val):
        return abs(val)

    @op(
        name="number-toByteString",
        input_type={"number": types.Number()},
        output_type=types.String(),
    )
    def to_byte_string(number):
        # Convert the number to a byte string.
        # JS version: String(numeral(inputs.in).format('0.00b'))
        unit_maxes = [
            (10**3, "B"),
            (10**6, "KB"),
            (10**9, "MB"),
            (10**12, "GB"),
            (10**15, "TB"),
            (10**18, "PB"),
            (10**21, "EB"),
            (10**24, "ZB"),
            (10**27, "YB"),
        ]
        unit = ""
        number_str = ""
        for unit_ndx, (unit_max, unit_str) in enumerate(unit_maxes):
            if number < unit_max or unit_ndx == len(unit_maxes) - 1:
                unit = unit_str
                if unit_ndx > 0:
                    reduced = number / unit_maxes[unit_ndx - 1][0]
                    number_str = f"{reduced:.2f}"
                else:
                    number_str = f"{number}"
                break

        return f"{number_str}{unit}"

    @op(
        name="number-toFixed",
        input_type={"val": types.Number(), "digits": types.Number()},
        output_type=types.Number(),
    )
    def to_fixed(val, digits):
        return round(val, int(digits))


def numbers_ops_output_type(input_types: dict[str, types.Type]) -> types.Type:
    arr_type = typing.cast(types.List, input_types["numbers"])
    if types.List(types.NoneType()).assign_type(arr_type):
        return types.NoneType()
    elif types.List(types.Number()).assign_type(arr_type):
        return arr_type.object_type
    else:
        return types.optional(arr_type.object_type)


def avg_output_type(input_types: dict[str, types.Type]) -> types.Type:
    arr_type = typing.cast(types.List, input_types["numbers"])
    if types.List(types.NoneType()).assign_type(arr_type):
        return types.NoneType()
    elif types.List(types.Number()).assign_type(arr_type):
        return types.Number()
    else:
        return types.optional(types.Number())


list_of_numbers_input_type = {"numbers": types.List(types.optional(types.Number()))}


@op(
    name="numbers-sum",
    input_type=list_of_numbers_input_type,
    output_type=numbers_ops_output_type,
)
def numbers_sum(numbers):
    numbers = [n for n in numbers if n != None]
    return sum(numbers) if len(numbers) > 0 else None


@op(
    name="numbers-avg",
    input_type=list_of_numbers_input_type,
    output_type=avg_output_type,
)
def numbers_avg(numbers):
    numbers = [n for n in numbers if n != None]
    return sum(numbers) / len(numbers) if len(numbers) > 0 else None


@op(
    name="numbers-min",
    input_type=list_of_numbers_input_type,
    output_type=numbers_ops_output_type,
)
def numbers_min(numbers):
    numbers = [n for n in numbers if n != None]
    return min(numbers) if len(numbers) > 0 else None


@op(
    name="numbers-max",
    input_type=list_of_numbers_input_type,
    output_type=numbers_ops_output_type,
)
def numbers_max(numbers):
    numbers = [n for n in numbers if n != None]
    return max(numbers) if len(numbers) > 0 else None


@op(
    name="numbers-argmax",
    input_type=list_of_numbers_input_type,
    output_type=numbers_ops_output_type,
)
def numbers_argmax(numbers):
    non_null_numbers = [n for n in numbers if n != None]
    if len(non_null_numbers) == 0:
        return None
    return numbers.index(max(non_null_numbers))


@op(
    name="numbers-argmin",
    input_type=list_of_numbers_input_type,
    output_type=numbers_ops_output_type,
)
def numbers_argmin(numbers):
    non_null_numbers = [n for n in numbers if n != None]
    if len(non_null_numbers) == 0:
        return None
    return numbers.index(min(non_null_numbers))


@op(
    name="numbers-stddev",
    input_type=list_of_numbers_input_type,
    output_type=numbers_ops_output_type,
)
def stddev(numbers):
    non_null_numbers = [n for n in numbers if n != None]
    if len(non_null_numbers) == 0:
        return None
    return np.std(non_null_numbers).tolist()


@op(
    name="number-random_normal",
)
def random_normal_single(mean: float, std: float) -> float:
    """
    Generate a single random number drawn from a univariate Gaussian distribution
    with mean `mean` and standard deviation `std`.

    Args:
        mean (float): The mean of the Gaussian distribution.
        std (float): The standard deviation of the Gaussian distribution.

    Returns:
        float: A single float representing the random number generated.
    """
    return random.gauss(mean, std)


@op(
    name="numbers-random_normal",
)
def random_normal(mean: float, std: float, n: int) -> list[float]:
    """
    Generate a list of `n` random numbers drawn from a univariate Gaussian distribution with
    mean `mean` and standard deviation `std`.

    Args:
        mean (float): The mean of the Gaussian distribution.
        std (float): The standard deviation of the Gaussian distribution.
        n (int): The number of random numbers to generate.

    Returns:
        list[float]: A list of `n` floats representing the random numbers generated.
    """
    samples = [random.gauss(mean, std) for _ in range(n)]
    return samples



import typing
import random
import datetime
import numpy as np

from . import context_state
from . import ref_util


def make_id() -> int:
    return random.randint(-(2**63), 2**63 - 1)


class HasBoxedRepr:
    val: typing.Any

    def __repr__(self):
        return f"<{self.__class__.__name__} ({self.val})>"


# Watch out, can't do "is None" on this!
# TODO: fix?
class BoxedNone(HasBoxedRepr):
    _id: typing.Optional[int] = None

    def __init__(self, val):
        self.val = val

    def __bool__(self):
        return bool(self.val)

    def __eq__(self, other):
        return self.val == other


def is_none(val):
    return val is None or isinstance(val, BoxedNone)


class BoxedBool(HasBoxedRepr):
    _id: typing.Optional[int] = None

    def __init__(self, val):
        self.val = val

    def __bool__(self):
        return self.val

    def __hash__(self):
        return hash(self.val)

    def __eq__(self, other):
        return self.val == other


class BoxedInt(int):
    _id: typing.Optional[int] = None


class BoxedFloat(float):
    _id: typing.Optional[int] = None


class BoxedStr(str):
    _id: typing.Optional[int] = None


class BoxedDict(dict):
    _id: typing.Optional[int] = None

    def _lookup_path(self, path: typing.List[str]):
        assert len(path) > 1
        edge_type = path[0]
        edge_path = path[1]
        assert edge_type == ref_util.DICT_KEY_EDGE_TYPE

        res = self[edge_path]
        remaining_path = path[2:]
        if remaining_path:
            return res._lookup_path(remaining_path)
        return res

    def __getitem__(self, __key: typing.Any) -> typing.Any:
        val = super().__getitem__(__key)
        return ref_util.val_with_relative_ref(
            self, val, [ref_util.DICT_KEY_EDGE_TYPE, str(__key)]
        )


class BoxedList(list):
    def _lookup_path(self, path: typing.List[str]):
        assert len(path) > 1
        edge_type = path[0]
        edge_path = path[1]
        assert edge_type == ref_util.LIST_INDEX_EDGE_TYPE

        res = self[int(edge_path)]
        remaining_path = path[2:]
        if remaining_path:
            return res._lookup_path(remaining_path)
        return res

    def __getitem__(self, __index: typing.Any) -> typing.Any:
        val = super().__getitem__(__index)
        return ref_util.val_with_relative_ref(
            self, val, [ref_util.LIST_INDEX_EDGE_TYPE, str(__index)]
        )


class BoxedDatetime(datetime.datetime):
    def __eq__(self, other):
        return (
            isinstance(other, datetime.datetime)
            and self.timestamp() == other.timestamp()
        )


class BoxedTimedelta(datetime.timedelta):
    def __eq__(self, other):
        return (
            isinstance(other, datetime.timedelta)
            and self.total_seconds() == other.total_seconds()
        )


def cannot_have_weakref(obj: typing.Any):
    return isinstance(obj, (BoxedInt, BoxedFloat, BoxedStr, BoxedBool, BoxedNone))


# See https://numpy.org/doc/stable/user/basics.subclassing.html
class BoxedNDArray(np.ndarray):
    def __new__(cls, input_array):
        obj = np.asarray(input_array).view(cls)
        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return


T = typing.TypeVar("T")


def box(
    obj: T,
) -> typing.Union[
    T,
    BoxedInt,
    BoxedFloat,
    BoxedStr,
    BoxedBool,
    BoxedDict,
    BoxedList,
    BoxedNDArray,
    BoxedNone,
    BoxedDatetime,
    BoxedTimedelta,
]:
    if type(obj) == int:
        return BoxedInt(obj)
    elif type(obj) == float:
        return BoxedFloat(obj)
    elif type(obj) == str:
        return BoxedStr(obj)
    elif type(obj) == bool:
        return BoxedBool(obj)
    elif type(obj) == dict:
        return BoxedDict(obj)
    elif type(obj) == list:
        return BoxedList(obj)
    elif type(obj) == np.ndarray:
        return BoxedNDArray(obj)
    elif type(obj) == datetime.datetime:
        return BoxedDatetime.fromtimestamp(obj.timestamp(), tz=datetime.timezone.utc)
    elif type(obj) == datetime.timedelta:
        return BoxedTimedelta(seconds=obj.total_seconds())
    elif obj is None:
        return BoxedNone(obj)
    return obj


def unbox(
    obj: T,
) -> typing.Union[
    T,
    int,
    float,
    str,
    bool,
    dict,
    list,
    np.ndarray,
    datetime.datetime,
    datetime.timedelta,
    None,
]:
    if type(obj) == BoxedInt:
        return int(obj)
    elif type(obj) == BoxedFloat:
        return float(obj)
    elif type(obj) == BoxedStr:
        return str(obj)
    elif type(obj) == BoxedBool:
        return bool(obj)
    elif type(obj) == BoxedDict:
        return dict(obj)
    elif type(obj) == BoxedList:
        return list(obj)
    elif type(obj) == BoxedNDArray:
        return np.array(obj)
    elif type(obj) == BoxedDatetime:
        return datetime.datetime.fromtimestamp(obj.timestamp())
    elif type(obj) == BoxedTimedelta:
        return datetime.timedelta(seconds=obj.total_seconds())
    elif type(obj) == BoxedNone:
        return None
    return obj


def is_boxed(obj: typing.Any) -> bool:
    return id(obj) == id(box(obj))



"""Basic autopatching of trackable libraries.

This module should not require any dependencies beyond the standard library. It should
check if libraries are installed and imported and patch in the case that they are.
"""


def autopatch_openai() -> None:
    try:
        import openai  # type: ignore
    except ImportError:
        pass
    else:
        if openai.__version__ < "1":
            print(
                "To automatically track openai calls, upgrade the openai package to a version >= '1.0'"
            )
            return
        from weave.monitoring.openai import patch

        patch()


def autopatch() -> None:
    autopatch_openai()



import dataclasses
import typing

import weave
from .. import panel
from .. import weave_internal


@weave.type()
class SliderConfig:
    min: weave.Node[float] = dataclasses.field(
        default_factory=lambda: weave_internal.make_const_node(weave.types.Float(), 0)
    )
    max: weave.Node[float] = dataclasses.field(
        default_factory=lambda: weave_internal.make_const_node(weave.types.Float(), 10)
    )
    step: weave.Node[float] = dataclasses.field(
        default_factory=lambda: weave_internal.make_const_node(weave.types.Float(), 0.1)
    )


@weave.type()
class Slider(panel.Panel):
    id = "Slider"
    config: typing.Optional[SliderConfig] = dataclasses.field(
        default_factory=SliderConfig
    )

    def __post_init__(self, *args):
        super().__post_init__(*args)
        if isinstance(self.input_node, weave.graph.VoidNode):
            self.__dict__["input_node"] = weave_internal.const(0)

    @weave.op()
    def value(self) -> float:
        return weave.use(self.input_node)



from ..decorator_arrow_op import arrow_op
from .. import weave_types as types
from ..ops_primitives import obj as primitives_obj

from ..arrow.list_ import ArrowWeaveList, ArrowWeaveListType


@arrow_op(
    name="ArrowWeaveListObject-__vectorizedGetattr__",
    input_type={
        "self": ArrowWeaveListType(types.Any()),
        "name": types.String(),
    },
    output_type=lambda input_types: ArrowWeaveListType(
        primitives_obj.getattr_output_type(
            {"self": input_types["self"].object_type, "name": input_types["name"]}
        )
    ),
    all_args_nullable=False,
)
def arrow_getattr(self, name):
    data = self._arrow_data
    t = types.non_none(self.object_type).property_types()[name]
    if types.optional(self.object_type):
        t = types.optional(t)
    return ArrowWeaveList(data.field(name), t, self._artifact)



from dataclasses import dataclass
import weave
from typing import Any, Callable, Optional
import numpy as np
from weave.weaveflow import Dataset, Model
from weave import op_def
from weave.weaveflow import util
import time
import inspect
import traceback

from rich.console import Console
from rich import print

console = Console()


def auto_summarize(data: weave.WeaveList) -> Optional[dict]:
    """Automatically summarize a WeaveList of (potentially nested) dicts.

    Will compute min/p25/avg/p75/max for all numeric columns.
    Will compute count and fraction for all boolean columns.
    Other leaf column types will be ignored.
    Also computes none_count and none_fraction for numeric and boolean columns.
    If a column is all None, result will be None

    Returns:
      dict of summary stats, with structure matching input dict structure.
    """
    if data.is_number():
        valid_data = [x for x in data if x is not None]
        if not valid_data:
            return None
        # Just avg and none_fraction for now. The others make the UI
        # too noisy. And all of these can be derived.
        return {
            # "min": float(np.min(valid_data)),
            # "p25": float(np.percentile(valid_data, 25)),
            "mean": float(np.mean(valid_data)),
            # "p75": float(np.percentile(valid_data, 75)),
            # "max": float(np.max(valid_data)),
            # "none_fraction": (len(data) - len(valid_data)) / len(data),
        }
    elif data.is_boolean():
        valid_data = [x for x in data if x is not None]
        count_true = valid_data.count(True)
        return {
            "true_count": count_true,
            "true_fraction": count_true / len(valid_data) if valid_data else 0,
            # "none_fraction": (len(data) - len(valid_data)) / len(data),
        }
    elif data.is_dict():
        result = {}
        for col_name in data.column_names:
            nested_data = data.column(col_name)
            summary = auto_summarize(nested_data)
            if summary is not None:
                result[col_name] = summary
        if not result:
            return None
        return result
    return None


@weave.type()
class Evaluation:
    dataset: Dataset
    scores: list[op_def.OpDef]

    # TODO: can't use regular callables here, Weave doesn't know about them.
    # example_to_model_input: Callable = lambda x: x["input"]
    example_to_model_input: op_def.OpDef

    @weave.op()
    async def predict_and_score(self, example: dict, model: Model) -> dict:
        model_input = self.example_to_model_input(example)
        try:
            prediction = await model.predict(model_input)
        except Exception as e:
            print("Prediction failed")
            traceback.print_exc()
            return {
                "prediction": None,
                "prediction_error": True,
                "scores": {},
                "score_errors": 0,
            }
        scores = {}
        score_errors = 0
        for scorer in self.scores:
            # TODO: if there are multiple of the same we need to distinguish
            scorer_name = scorer.common_name
            try:
                result = scorer(example, prediction)
                if inspect.iscoroutine(result):
                    result = await result
            except Exception as e:
                print(f"Score {scorer_name} exception")
                traceback.print_exc()
                result = None
                score_errors += 1
            scores[scorer_name] = result

        return {
            "prediction": prediction,
            "prediction_error": False,
            "scores": scores,
            "score_errors": score_errors,
        }

    @weave.op()
    async def summarize(self, eval_table: weave.WeaveList) -> dict:
        summary = {}
        prediction_summary = auto_summarize(eval_table.column("prediction"))
        if prediction_summary:
            summary["prediction"] = prediction_summary
        for scorer in self.scores:
            scorer_name = scorer.common_name
            scorer_scores = eval_table.column("scores").column(scorer_name)
            summary[scorer_name] = auto_summarize(scorer_scores)  # type: ignore
        return summary

    @weave.op()
    async def evaluate(self, model: Model) -> dict:
        eval_rows = []

        start_time = time.time()

        async def eval_example(example: dict) -> dict:
            eval_row = await self.predict_and_score(example, model)
            return eval_row

        prediction_errors = 0
        score_errors = 0
        n_complete = 0
        with console.status("Evaluating...") as status:
            _rows = self.dataset.rows
            async for example, eval_row in util.async_foreach(_rows, eval_example, 30):
                n_complete += 1
                prediction_errors += int(eval_row["prediction_error"])
                score_errors += eval_row["score_errors"]
                duration = time.time() - start_time
                status.update(
                    f"Evaluating... {duration:.2f}s [{n_complete} / {len(self.dataset.rows)} complete] [{prediction_errors} prediction errors] [{score_errors} score errors]"
                )
                eval_rows.append(eval_row)

        with console.status("Summarizing...") as status:
            eval_table: weave.WeaveList = weave.WeaveList(eval_rows)

            summary = await self.summarize(eval_table)

        return summary



import typing


import contextvars
import contextlib

from . import engine_trace

statsd = engine_trace.statsd()  # type: ignore

_memo_storage: contextvars.ContextVar[typing.Optional[dict]] = contextvars.ContextVar(
    "memo_storage", default=None
)


@contextlib.contextmanager
def memo_storage() -> typing.Generator[None, None, None]:
    # Here we only construct a new context if one does not already exist.
    # This is because we often have recursive calls to the executor (particularly
    # in the case of refinement during querying). In such cases we do not want
    # to reset the memoization storage, as this would cause us to lose the
    # memoized values from the previous call! We might want to do something
    # similar for other context variables.
    token = None
    if _memo_storage.get() is None:
        token = _memo_storage.set({})
    try:
        yield
    finally:
        if token is not None:
            _memo_storage.reset(token)


class NoValue:
    pass


NO_VALUE = NoValue()


def memo(f: typing.Any) -> typing.Any:
    def call_memo(*args: typing.Any, **kwargs: typing.Any) -> typing.Any:
        storage = _memo_storage.get()
        if storage is None:
            return f(*args, **kwargs)
        key = (f, args, tuple(kwargs.items()))
        try:
            val = storage[key]
            # statsd.increment("weave.memo.hit")
            return val
        except KeyError:
            pass
        # statsd.increment("weave.memo.miss")
        result = f(*args, **kwargs)
        storage[key] = result
        return result

    return call_memo



import textwrap

from .weave_types import *
from .language_features.tagging import tagged_value_type


def short_type(type: Type) -> str:
    s = type.__repr__()
    if len(s) > 40:
        s = s[:40] + "..."
    return f"<{s}>"


def why_not_assignable(to_type: Type, from_type: Type) -> typing.Optional[str]:
    """Returns None if assignable, otherwise a string explaining why not.

    This is the start of a the implementation, however, it is not yet complete.
    """
    reasons: list[str] = []
    if to_type == Any() or from_type == Any():
        return None
    elif isinstance(from_type, UnionType):
        for from_member in from_type.members:
            reason = why_not_assignable(to_type, from_member)
            if reason is not None:
                reasons.append(reason)
    elif isinstance(to_type, UnionType):
        for to_member in to_type.members:
            reason = why_not_assignable(to_member, from_type)
            if reason is None:
                return None
            reasons.append(f"{short_type(to_member)}: {reason}")

    elif isinstance(from_type, tagged_value_type.TaggedValueType) and not isinstance(
        to_type, tagged_value_type.TaggedValueType
    ):
        reason = why_not_assignable(to_type, from_type.value)
        if reason is not None:
            reasons.append(reason)

    elif isinstance(from_type, Const) and not isinstance(to_type, Const):
        reason = why_not_assignable(to_type, from_type.val_type)
        if reason is not None:
            reasons.append(reason)

    elif not isinstance(from_type, Const) and isinstance(to_type, Const):
        # Weird const behavior, allow assigning non-const to const (see weave_types)
        reason = why_not_assignable(to_type.val_type, from_type)
        if reason is not None:
            reasons.append(reason)

    elif isinstance(from_type, Function) and not isinstance(to_type, Function):
        reason = why_not_assignable(to_type, from_type.output_type)
        if reason is not None:
            reasons.append(reason)

    elif isinstance(from_type, TypedDict) and isinstance(to_type, Dict):
        for k, from_td_type in from_type.property_types.items():
            sub_reason = why_not_assignable(to_type.object_type, from_td_type)
            if sub_reason is not None:
                reasons.append(
                    f"Property {k} is not assignable\n{textwrap.indent(sub_reason, '  ')}"
                )

    elif isinstance(from_type, TypedDict) and isinstance(to_type, TypedDict):
        for k, to_type_type in to_type.property_types.items():
            from_type_type = from_type.property_types.get(k)
            if from_type_type is None:
                reasons.append(f"Missing property {k}")
            else:
                sub_reason = why_not_assignable(to_type_type, from_type_type)
                if sub_reason is not None:
                    reasons.append(
                        f"Property {k} is not assignable\n{textwrap.indent(sub_reason, '  ')}"
                    )

    elif to_type.name == from_type.name:
        type_vars = to_type.type_vars
        for k, to_type_type in type_vars.items():
            from_type_type = getattr(from_type, k)
            if from_type_type is None:
                reasons.append(f"Missing property {k}")
            else:
                sub_reason = why_not_assignable(to_type_type, from_type_type)
                if sub_reason is not None:
                    reasons.append(
                        f"Property {k} is not assignable\n{textwrap.indent(sub_reason, '  ')}"
                    )
    else:
        reasons.append("Incompatible types")
    if reasons:
        indented_reasons = textwrap.indent("\n".join(reasons), "  ")
        return f"{short_type(to_type)} !<- {short_type(from_type)}\n{indented_reasons}"

    # Check if our implementation above matches the actual assignability
    # implementation.
    is_assignable = to_type.assign_type(from_type)
    if reasons and is_assignable:
        print(
            "why_not_assignable programming error. reasons but assignable",
            to_type,
            from_type,
            reasons,
        )
    elif not reasons and not is_assignable:
        print(
            "why_not_assignable programming error. no reasons but not assignable",
            to_type,
            from_type,
            reasons,
        )

    return None



from weave import weave_types
from ..arrow.arrow import ArrowWeaveListType
from ..core_types import StreamTableType
from ..api import op
from .project_ops import project
from .. import compile
from .. import op_def


def _get_history_node(stream_table: StreamTableType):
    with op_def.no_refine():
        return (
            project(stream_table.entity_name, stream_table.project_name)
            .run(stream_table.table_name)
            .history3()
            .dropTags()
        )


@op()
def _rows_type(stream_table: StreamTableType) -> weave_types.Type:
    with compile.enable_compile():
        return compile.compile([_get_history_node(stream_table)])[0].type


@op(
    name="stream_table-rows",
    output_type=ArrowWeaveListType(weave_types.TypedDict({})),
    refine_output_type=_rows_type,
)
def rows(stream_table: StreamTableType):
    return _get_history_node(stream_table)



from .. import context_state as _context_state

_loading_builtins_token = _context_state.set_loading_built_ins()

from .any import *
from .list_ import *
from .obj import *
from .type import *
from .string import *
from .number import *
from .number_bin import *
from .boolean import *
from .date import *
from .dict import *
from .timestamp_bin import *
from .set_ import *
from .file import *
from .file_local import *
from .file_artifact import *
from .weave_api import *
from .pandas_ import *
from .csv_ import *
from .json_ import *
from .sql import *
from .artifacts import *
from .random_junk import *
from .image import *
from .html import *
from .markdown import *
from .op_def import *
from ..language_features.tagging.tagging_ops import *
from .list_tag_getters import *
from .geom import *
from .server import *

_context_state.clear_loading_built_ins(_loading_builtins_token)



import aioprocessing
import janus
from typing import Optional, Generic, TypeVar

QueueItemType = TypeVar("QueueItemType")


class Queue(Generic[QueueItemType]):
    async def async_put(self, item: QueueItemType) -> None:
        raise NotImplementedError

    async def async_get(
        self,
    ) -> QueueItemType:
        raise NotImplementedError

    async def async_join(self) -> None:
        raise NotImplementedError

    def task_done(self) -> None:
        raise NotImplementedError

    def put(self, item: QueueItemType) -> None:
        raise NotImplementedError

    def get(self, block: bool = True, timeout: Optional[int] = None) -> QueueItemType:
        raise NotImplementedError

    def join(self) -> None:
        raise NotImplementedError


class ProcessQueue(Queue, Generic[QueueItemType]):
    def __init__(self, maxsize: int = 0):
        self._queue: aioprocessing.Queue = aioprocessing.AioJoinableQueue(
            maxsize=maxsize
        )

    async def async_put(self, item: QueueItemType) -> None:
        await self._queue.coro_put(item)

    async def async_get(
        self,
    ) -> QueueItemType:
        return await self._queue.coro_get()

    async def async_join(self) -> None:
        await self._queue.coro_join()

    def put(self, item: QueueItemType) -> None:
        self._queue.put(item)

    def get(self, block: bool = True, timeout: Optional[int] = None) -> QueueItemType:
        return self._queue.get(block=block, timeout=timeout)

    def join(self) -> None:
        self._queue.join()

    def task_done(self) -> None:
        self._queue.task_done()


class ThreadQueue(Queue, Generic[QueueItemType]):
    @property
    def queue(self) -> janus.Queue:
        if not hasattr(self, "_queue"):
            self.init()
        return self._queue

    def __init__(
        self,
        maxsize: int = 0,
    ):
        self.maxsize = maxsize

    async def async_put(self, item: QueueItemType) -> None:
        await self.queue.async_q.put(item)

    async def async_get(self) -> QueueItemType:
        return await self.queue.async_q.get()

    async def async_join(self) -> None:
        await self.queue.async_q.join()

    def put(self, item: QueueItemType) -> None:
        self.queue.sync_q.put(item)

    def get(self, block: bool = True, timeout: Optional[int] = None) -> QueueItemType:
        return self.queue.sync_q.get(block=block, timeout=timeout)

    def join(self) -> None:
        self.queue.sync_q.join()

    def task_done(self) -> None:
        self.queue.sync_q.task_done()

    def init(self) -> None:
        self._queue: janus.Queue = janus.Queue(maxsize=self.maxsize)



# This is deprecated in favor of the new wandb_api.py module.
# TODO: remove uses of this and delete.

from wandb.apis import public
from wandb.apis.public.api import gql
from wandb.sdk.internal.internal_api import _thread_local_api_settings
import logging
import typing

from wandb.errors import CommError as WandbCommError

from . import errors

import graphql
from graphql import GraphQLSchema

from requests import exceptions


def wandb_public_api() -> public.Api:
    return public.Api(timeout=30)


def assert_wandb_authenticated() -> None:
    authenticated = (
        wandb_public_api().api_key is not None
        or _thread_local_api_settings.cookies is not None
    )
    if not authenticated:
        raise errors.WeaveWandbAuthenticationException(
            "Unable to log data to W&B. Please authenticate by setting WANDB_API_KEY or running `wandb init`."
        )


def query_with_retry(
    query_str: str,
    variables: dict[str, typing.Any] = {},
    num_timeout_retries: int = 0,
) -> typing.Any:
    if num_timeout_retries < 0:
        raise ValueError("num_timeout_retries must be >= 0")
    for attempt_no in range(num_timeout_retries + 1):
        try:
            return wandb_public_api().client.execute(
                gql(query_str),
                variable_values=variables,
            )
        except exceptions.Timeout as e:
            if attempt_no == num_timeout_retries:
                raise
            logging.warn(
                f'wandb GQL query timed out: "{e}", retrying (num_attempts={attempt_no + 1})'
            )


def introspect_server_schema(num_timeout_retries: int = 0) -> GraphQLSchema:
    introspection_query = graphql.get_introspection_query()
    payload = query_with_retry(introspection_query, {}, num_timeout_retries)
    return graphql.build_client_schema(payload)


def wandb_gql_query(
    query_str: str,
    variables: dict[str, typing.Any] = {},
    num_timeout_retries: int = 0,
) -> typing.Any:
    return query_with_retry(query_str, variables, num_timeout_retries)


def set_wandb_thread_local_api_settings(
    api_key: typing.Optional[str],
    cookies: typing.Optional[typing.Dict],
    headers: typing.Optional[typing.Dict],
) -> None:
    _thread_local_api_settings.api_key = api_key
    _thread_local_api_settings.cookies = cookies
    _thread_local_api_settings.headers = headers


class WandbThreadLocalApiSettings(typing.TypedDict):
    api_key: typing.Optional[str]
    cookies: typing.Optional[dict]
    headers: typing.Optional[dict]


def reset_wandb_thread_local_api_settings() -> None:
    _thread_local_api_settings.api_key = None
    _thread_local_api_settings.cookies = None
    _thread_local_api_settings.headers = None


_WANDB_CLIENT_PATCHED = False


def do_patch() -> None:
    global _WANDB_CLIENT_PATCHED
    if _WANDB_CLIENT_PATCHED:
        return

    ### Start _parse_artifact_path Patch ###
    # Can be removed after https://github.com/wandb/wandb/pull/6083 is merged and deployed.
    orig_parse_artifact_path = public.Api._parse_artifact_path

    def new_parse_artifact_path(self, path):  # type: ignore
        # Adding this short circuit skips the potential access of
        # self.default_entity which incurs a network call. However, if the path
        # is fully qualified, then this `entity` is thrown away.
        parts = [] if path is None else path.split("/")
        if len(parts) == 3:
            return parts
        return orig_parse_artifact_path(self, path)

    public.Api._parse_artifact_path = new_parse_artifact_path  # type: ignore
    ### End _parse_artifact_path Patch ###

    _WANDB_CLIENT_PATCHED = True


do_patch()



from . import weave_types as types
from . import weave_internal

# TODO: Consider if this should accept a *args, **kwargs?
def make(cls, kwargs={}):
    args = {types.to_weavejs_typekey(k): v for k, v in kwargs.items()}
    args["name"] = weave_internal.make_const_node(types.String(), cls.class_type_name())
    # TODO: Should we define this Op as a native Python op?
    return weave_internal.make_output_node(types.TypeType(), "type-__newType__", args)


types.Type._make = make  # type: ignore



import weave
from weave.ecosystem import torchvision, torch_mnist_model_example


def test_mnist_notebook():
    # Cell 1:
    mnist_dataset = torchvision.mnist(
        100
    )  # Or try mnist.food101(100), but currently broken
    weave.use(mnist_dataset)

    # Cell 2:
    hyperparams = {
        "fc_layer_size": 256,
        "dropout": 0.5,
        "epochs": 5,
        "learning_rate": 0.005,
        "batch_size": 128,
    }
    train_split = mnist_dataset["data"]["train"]
    # train_split.pick('image')
    model = torch_mnist_model_example.train(
        train_split.pick("image"), train_split.pick("label"), hyperparams
    )

    # Cell 3:
    test_split_images = mnist_dataset["data"]["test"].pick("image")
    preds = model.predict(test_split_images)
    weave.use(preds)



import torch
import weave


class TorchTensorType(weave.types.Type):
    instance_classes = torch.Tensor
    instance_class = torch.Tensor

    def save_instance(self, obj, artifact, name):
        with artifact.new_file(f"{name}.pt", binary=True) as f:
            torch.save(obj, f)

    def load_instance(self, artifact, name, extra=None):
        with artifact.open(f"{name}.pt", binary=True) as f:
            return torch.load(f)


class TorchModelType(weave.types.Type):
    instance_classes = torch.nn.Sequential
    instance_class = torch.nn.Sequential

    def save_instance(self, obj, artifact, name):
        with artifact.new_file(f"{name}.pt", binary=True) as f:
            torch.save(obj, f)

    def load_instance(self, artifact, name, extra=None):
        with artifact.open(f"{name}.pt", binary=True) as f:
            return torch.load(f)



import urllib
import json
from wandb import util as wb_util

from . import environment
from . import context_state

BROWSE3_PATH = "browse3"
WORKSPACE_SLUG = "weaveflow"


def remote_project_root_url(entity_name: str, project_name: str) -> str:
    return f"{wb_util.app_url(environment.wandb_base_url())}/{entity_name}/{project_name}/{WORKSPACE_SLUG}"


def local_project_root_url(entity_name: str, project_name: str) -> str:
    return f"http://localhost:3000/{BROWSE3_PATH}/{entity_name}/{project_name}"


def project_root_url(entity_name: str, project_name: str) -> str:
    if context_state._use_local_urls.get():
        return local_project_root_url(entity_name, project_name)
    else:
        return remote_project_root_url(entity_name, project_name)


def project_path(entity_name: str, project_name: str) -> str:
    return f"{project_root_url(entity_name, project_name)}"


def op_version_path(
    entity_name: str, project_name: str, op_name: str, op_version: str
) -> str:
    return f"{project_root_url(entity_name, project_name)}/ops/{op_name}/versions/{op_version}"


def object_version_path(
    entity_name: str, project_name: str, object_name: str, op_version: str
) -> str:
    return f"{project_root_url(entity_name, project_name)}/objects/{object_name}/versions/{op_version}"


def call_path(entity_name: str, project_name: str, call_id: str) -> str:
    return f"{project_root_url(entity_name, project_name)}/calls/{call_id}"


def call_path_as_peek(entity_name: str, project_name: str, call_id: str) -> str:
    return f"{call_path(entity_name, project_name, call_id)}?convertToPeek=true"


def use_local_urls() -> None:
    context_state._use_local_urls.set(True)


def use_remote_urls() -> None:
    context_state._use_local_urls.set(False)



from .. import context_state as _context_state

_loading_builtins_token = _context_state.set_loading_built_ins()

from ..arrow.arrow import *
from ..arrow.list_ import *
from ..arrow.convert import *
from .vectorize import vectorize
from . import ops
from .list_range import range

_context_state.clear_loading_built_ins(_loading_builtins_token)



"""Monitoring & tracing"""

import asyncio
import contextlib
import contextvars
import dataclasses
import datetime
import inspect
import sys
import typing
import uuid
import logging


from ..wandb_interface.wandb_stream_table import StreamTable
from .. import errors
from .. import graph
from .. import stream_data_interfaces
from .. import graph_client_context
from .. import graph_client_wandb_art_st
from .. import run_context
from .. import run_streamtable_span

logger = logging.getLogger(__name__)

_global_monitor: typing.Optional["Monitor"] = None

_attributes: contextvars.ContextVar[
    typing.Dict[str, typing.Any]
] = contextvars.ContextVar("_attributes", default={})


# Matches OpenTelemetry
class StatusCode:
    UNSET = "UNSET"
    SUCCESS = "SUCCESS"
    ERROR = "ERROR"


def _arguments_to_dict(
    fn: typing.Callable, args: tuple[typing.Any, ...], kwargs: dict[str, typing.Any]
) -> dict[str, typing.Any]:
    """Given a function and arguments used to call that function, return a dictionary of key value pairs
    that are the bound arguments to the function.
    """
    sig = inspect.signature(fn)
    bound_params = sig.bind(*args, **kwargs)
    bound_params.apply_defaults()
    got_args = dict(bound_params.arguments)
    return got_args


def get_function_name(fn: typing.Callable) -> str:
    module_name = fn.__module__

    # Check if the function is bound to a class
    if inspect.ismethod(fn):
        class_name = fn.__self__.__class__.__name__
        return f"{module_name}.{class_name}.{fn.__name__}"
    else:
        return f"{module_name}.{fn.__name__}"


class Span:
    # When this is None, monitor is disabled
    _streamtable: typing.Optional[StreamTable]

    # These are OpenTelemetry standard

    parent_id: typing.Optional[str]
    trace_id: str
    id: str  # span_id
    name: str
    status_code: str = StatusCode.UNSET
    start_time: datetime.datetime
    end_time: typing.Optional[datetime.datetime]

    # OpenTelemetry does not support these.

    # Arbitrary key/value pairs. Values can be any Weave type, as
    # opposed to opentelemetry where values can only be scalars or
    # lists of homoegenous scalar types.
    attributes: dict[str, typing.Any]
    # The parameters to the function being traced.
    inputs: typing.Optional[typing.Dict[str, typing.Any]]
    # The output of the function being traced.
    output: typing.Optional[typing.Any]
    # If an exception occurred, this will be set.
    exception: typing.Optional[Exception]
    # Typically summary should contain metrics about resources used
    # during the function's execution (cpu, disk, tokens, api cost etc)
    # or other information created in the course of actually executing
    # the function.
    summary: dict[str, typing.Any]

    _autoclose: bool = True

    def __init__(
        self,
        name: str,
        _streamtable: typing.Optional[StreamTable] = None,
        parent_id: typing.Optional[str] = None,
        trace_id: typing.Optional[str] = None,
        attributes: typing.Optional[dict[str, typing.Any]] = None,
        inputs: typing.Optional[typing.Dict[str, typing.Any]] = None,
        output: typing.Optional[typing.Any] = None,
        summary: typing.Optional[dict[str, typing.Any]] = None,
    ):
        self.name = name
        self._streamtable = _streamtable
        self.parent_id = parent_id
        if trace_id is None:
            trace_id = str(uuid.uuid4())
        self.trace_id = trace_id
        if attributes is None:
            attributes = {}
        self.attributes = attributes
        self.status_code = StatusCode.UNSET
        self.span_id = str(uuid.uuid4())
        self.start_time = datetime.datetime.now()
        self.end_time = None
        self.inputs = inputs
        self.output = output
        self.exception = None
        if summary is None:
            summary = {}
        self.summary = summary

    def close(self) -> None:
        if self.status_code == StatusCode.UNSET:
            self.status_code = StatusCode.SUCCESS
        self.end_time = datetime.datetime.now()
        if self._streamtable is not None:
            self._streamtable.log(self.asdict())

    def disable_autoclose(self) -> None:
        self._autoclose = False

    def autoclose(self) -> None:
        if self._autoclose:
            self.close()

    def asdict(self) -> stream_data_interfaces.TraceSpanDict:
        if self.end_time is None:
            raise ValueError(
                "Cannot log a span that has not been ended.  Call span.end() before logging."
            )
        start_time_s = self.start_time.timestamp()
        end_time_s = self.end_time.timestamp()
        self.summary["latency_s"] = end_time_s - start_time_s
        return {
            "parent_id": self.parent_id,
            "trace_id": self.trace_id,
            "span_id": self.span_id,
            "name": self.name,
            "status_code": self.status_code,
            "start_time_s": start_time_s,
            "end_time_s": end_time_s,
            "inputs": self.inputs,
            "output": self.output,
            "exception": (
                f"{type(self.exception).__name__}: {str(self.exception)}"
                if self.exception is not None
                else None
            ),
            "attributes": self.attributes,
            "summary": self.summary,
        }

    def asdict_unsafe(self) -> stream_data_interfaces.TraceSpanDict:
        start_time_s = self.start_time.timestamp()
        end_time_s = None
        if self.end_time is not None:
            end_time_s = self.end_time.timestamp()
            self.summary["latency_s"] = end_time_s - start_time_s
        return {
            "parent_id": self.parent_id,
            "trace_id": self.trace_id,
            "span_id": self.span_id,
            "name": self.name,
            "status_code": self.status_code,
            "start_time_s": start_time_s,
            "end_time_s": end_time_s,  # type: ignore
            "inputs": self.inputs,
            "output": self.output,
            "exception": (
                f"{type(self.exception).__name__}: {str(self.exception)}"
                if self.exception is not None
                else None
            ),
            "attributes": self.attributes,
            "summary": self.summary,
        }


# A type used to indicate that inputs is guaranteed to be set, for pre and post process callbacks.
class SpanWithInputs(Span):
    inputs: typing.Dict[str, typing.Any]


class SpanWithInputsAndOutput(Span):
    inputs: typing.Dict[str, typing.Any]
    output: typing.Any


@dataclasses.dataclass
class Monitor:
    # Can be init'd with a specific streamtable
    _streamtable: typing.Optional[StreamTable]

    _showed_not_logging_warning: bool = False

    @property
    def streamtable(self) -> typing.Optional[StreamTable]:
        if self._streamtable:
            return self._streamtable
        # If we weren't init'd with a streamtable, try to get the global
        # one.
        client = graph_client_context.get_graph_client()
        if client:
            if isinstance(
                client, graph_client_wandb_art_st.GraphClientWandbArtStreamTable
            ):
                self._streamtable = client.runs_st
            else:
                # TODO: we need to refactor monitor to make use of graph_client
                print("Low-level tracing only works with wandb client currently")
        return self._streamtable

    @contextlib.contextmanager
    def span(self, name: str) -> typing.Iterator[Span]:
        if not self._showed_not_logging_warning and self.streamtable is None:
            self._showed_not_logging_warning = True
            print(
                "WARNING: Not logging spans.  Call weave.monitor.init_monitor() to enable logging."
            )

        # A song and dance to switch between span/run semantics. Will be cleaned
        # up as we switch entirely to the run interface.

        parent_run = run_context.get_current_run()
        trace_id = None
        if parent_run is not None:
            parent_id = parent_run.id
            trace_id = parent_run.trace_id
        else:
            parent_id = None
        span = Span(name, self.streamtable, parent_id, trace_id, _attributes.get())
        run = run_streamtable_span.RunStreamTableSpan(span.asdict_unsafe())
        with run_context.current_run(run):
            try:
                yield span
            finally:
                span.autoclose()

    @contextlib.contextmanager
    def attributes(self, attributes: typing.Dict[str, typing.Any]) -> typing.Iterator:
        cur_attributes = {**_attributes.get()}
        cur_attributes.update(attributes)

        token = _attributes.set(cur_attributes)
        try:
            yield
        finally:
            _attributes.reset(token)

    def trace(
        self,
        *,
        static_attributes: typing.Optional[typing.Dict[str, typing.Any]] = None,
        preprocess: typing.Optional[typing.Callable] = None,
        postprocess: typing.Optional[typing.Callable] = None,
    ) -> typing.Callable[..., typing.Callable[..., typing.Any]]:
        def decorator(fn: typing.Callable[..., typing.Any]) -> typing.Any:
            if asyncio.iscoroutinefunction(fn):

                async def async_wrapper(
                    *args: typing.Any, **kwargs: typing.Any
                ) -> typing.Any:
                    attributes = {}
                    attributes.update(static_attributes or {})
                    attributes.update(kwargs.pop("monitor_attributes", {}))
                    with self.attributes(attributes):
                        with self.span(get_function_name(fn)) as span:
                            span.inputs = _arguments_to_dict(fn, args, kwargs)
                            try:
                                span.output = await fn(*args, **kwargs)
                            except Exception as e:
                                span.status_code = StatusCode.ERROR
                                span.exception = e
                                raise
                            return span.output

                return async_wrapper

            else:

                def sync_wrapper(*args: typing.Any, **kwargs: typing.Any) -> typing.Any:
                    attributes = {}
                    attributes.update(static_attributes or {})
                    attributes.update(kwargs.pop("monitor_attributes", {}))
                    with self.attributes(attributes):
                        with self.span(get_function_name(fn)) as span:
                            span.inputs = _arguments_to_dict(fn, args, kwargs)
                            if preprocess:
                                preprocess(span)
                            try:
                                span.output = fn(*args, **kwargs)
                                output = span.output
                                if postprocess:
                                    output = postprocess(span)
                            except Exception as e:
                                span.status_code = StatusCode.ERROR
                                span.exception = e
                                raise
                            return output

                sync_wrapper.__name__ = fn.__name__
                sync_wrapper.__doc__ = fn.__doc__

                return sync_wrapper

        return decorator

    def rows(self) -> typing.Optional[graph.Node]:
        if self.streamtable is None:
            return None
        return self.streamtable.rows()


def _init_monitor_streamtable(stream_key: str) -> typing.Optional[StreamTable]:
    tokens = stream_key.split("/")
    if len(tokens) == 2:
        project_name, stream_name = tokens
        entity_name = None
    elif len(tokens) == 3:
        entity_name, project_name, stream_name = tokens
    else:
        raise ValueError(
            "stream_key must be of the form 'entity/project/stream_name' or 'project/stream_name'"
        )

    stream_table = None
    try:
        stream_table = StreamTable(
            table_name=stream_name,
            project_name=project_name,
            entity_name=entity_name,
        )
    except errors.WeaveWandbAuthenticationException:
        logger.error("Monitoring disabled because WANDB_API_KEY is not set.")
        print("Couldn't find W&B API key, disabling monitoring.", file=sys.stderr)
        print(
            "Set the WANDB_API_KEY env variable to enable monitoring.",
            file=sys.stderr,
        )
    return stream_table


def default_monitor() -> Monitor:
    """Get the global Monitor."""
    global _global_monitor
    if _global_monitor is None:
        _global_monitor = Monitor(None)
    return _global_monitor


def _get_global_monitor() -> typing.Optional[Monitor]:
    client = graph_client_context.get_graph_client()
    if client is not None:
        if not isinstance(
            client, graph_client_wandb_art_st.GraphClientWandbArtStreamTable
        ):
            raise ValueError(
                "monitor logging (via openai patch for example) is only supported with wandb client currently"
            )
        return Monitor(client.runs_st)
    return _global_monitor


def new_monitor(stream_key: str) -> Monitor:
    """Create a new Monitor"""
    return Monitor(_init_monitor_streamtable(stream_key))


def init_monitor(stream_key: str) -> Monitor:
    """Initialize the global monitor and return it."""
    global _global_monitor
    client = graph_client_context.get_graph_client()
    if client:
        raise ValueError("weave.init already called, init_monitor is invalid.")
    stream_table = _init_monitor_streamtable(stream_key)
    if _global_monitor is None:
        _global_monitor = Monitor(stream_table)
    else:
        _global_monitor._streamtable = stream_table
    return _global_monitor


def deinit_monitor() -> None:
    global _global_monitor
    _global_monitor = None


def trace(
    preprocess: typing.Optional[typing.Callable] = None,
    postprocess: typing.Optional[typing.Callable] = None,
) -> typing.Callable[..., typing.Callable[..., typing.Any]]:
    return default_monitor().trace(preprocess=preprocess, postprocess=postprocess)



import faiss
import typing
import weave

from weave import artifact_fs


class FaissIndexType(weave.types.Type):
    instance_classes = [faiss.Index]

    def save_instance(
        self, obj: faiss.Index, artifact: artifact_fs.FilesystemArtifact, name: str
    ) -> None:
        with artifact.writeable_file_path(f"{name}.faissindex") as write_path:
            faiss.write_index(obj, write_path)

    def load_instance(
        self,
        artifact: artifact_fs.FilesystemArtifact,
        name: str,
        extra: typing.Optional[list[str]] = None,
    ) -> faiss.Index:
        return faiss.read_index(artifact.path(f"{name}.faissindex"))



import typing
import weave

from .dataset import Dataset


@weave.type()
class ChatModel:
    @weave.op()
    def complete(self, messages: typing.Any) -> typing.Any:
        ...

    @weave.op()
    def finetune(
        self,
        train_dataset: Dataset,
        valid_dataset: Dataset,
        hyperparameters: typing.Any,
    ) -> typing.Any:
        from .openai import OpenaiChatModel
        from .anyscale import AnyscaleChatModel

        if not isinstance(self, (OpenaiChatModel, AnyscaleChatModel)):
            # This is because the following code assumes instance properties
            # that are not defined on the base class. Hack for weaveflow merge
            raise ValueError(
                "finetuning implementation only supported for OpenAI and Anyscale models"
            )

        import os
        from openai import OpenAI

        client = OpenAI(
            base_url=self.base_url, api_key=os.environ[self.api_key_env_var]
        )
        import time
        import json
        from io import BytesIO

        train_str = BytesIO()
        valid_str = BytesIO()

        for row in train_dataset.rows:
            train_str.write((json.dumps(row) + "\n").encode())

        for row in valid_dataset.rows:
            valid_str.write((json.dumps(row) + "\n").encode())

        train_str.seek(0)
        valid_str.seek(0)

        training_file_id = client.files.create(file=train_str, purpose="fine-tune").id
        print("Training file ID", training_file_id)

        valid_file_id = client.files.create(file=valid_str, purpose="fine-tune").id
        print("Valid file ID", valid_file_id)

        finetuning_job_id = client.fine_tuning.jobs.create(
            training_file=training_file_id,
            validation_file=valid_file_id,
            hyperparameters=hyperparameters,
            model=self.model_name,
        ).id
        print("Fine-tuning job ID", finetuning_job_id)

        while True:
            # This should be tracked with an async run
            fine_tune_result = client.fine_tuning.jobs.retrieve(
                finetuning_job_id,
            )

            print("RESULT", fine_tune_result)
            if (
                fine_tune_result.finished_at != None
                and fine_tune_result.fine_tuned_model != None
            ):
                break
            time.sleep(5)

        return self.__class__(fine_tune_result.fine_tuned_model)  # type: ignore



import typing
import copy

from . import graph
from . import ref_base
from . import weave_types as types

# Notes for the future:
# - I added list.lookup to lookup rows in a list by ID. I think we probably should
# add a weave.Table type which is a list of dicts that have ID. Then we can define
# node reference to use id lookup instead of index. Maybe figure out how to make
# it a multi-index so we can match pandas.


def node_to_ref(node: graph.Node) -> typing.Optional[ref_base.Ref]:
    """Converts a Node to equivalent Ref if possible.

    Specific call sequences can be converted to refs. For example:
      get(<object_path>)[0]['a'] can be converted to a Ref.

    The rule is:
      - a get call can be converted to a Ref
      - any __getitem__ (or pick) call can be converted to a Ref if its input is a Node
        that can be converted to a Ref

    It's desirable to store these as Refs rather than nodes because:
      - The representation is much more compact
      - A ref is a guarantee of existence
      - Artifacts is aware of cross-artifact references and keeps them
        consistent (it does not allow deleting an artifact that has existing
          depending artifacts)

    This is used when saving objects, to achieve cross-artifact references.

    A user may compose objects from values fetched from other objects. For example:
    x = weave.save({'a': 5})
    y = weave.save({'b': x['a']})

    Since Weave is lazy, x['a'] produces a Node. When we save objects, if a Node
      is encountered we try to convert it to a Ref using this function. If that
      succeeds, we save in Ref format instead of Node format.

    # TODO: we need to tell the artifact this Ref exists, so that it creates
    #   the database dependency, which is how we achieve consistency.
    """
    nodes = graph.linearize(node)
    if nodes is None:
        return None

    # First node in chain must be get op
    # TODO: check it's a get of a specific artifact version, not an alias!
    #     maybe we should have two different ops for that, so we can distinguish
    #     via type instead of via string checking.
    if nodes[0].from_op.name != "get":
        return None
    get_arg0 = list(nodes[0].from_op.inputs.values())[0]
    if not isinstance(get_arg0, graph.ConstNode):
        return None
    uri = get_arg0.val
    ref = ref_base.Ref.from_str(uri)
    if ref.extra is None:
        ref.extra = []

    # Remaining nodes in chain must be __getitem__
    for node in nodes[1:]:
        if not (
            node.from_op.name.endswith("__getitem__")
            # Allow pick too, but this is kinda busted. What if both exist?
            # I think we should maybe just use __getitem__ to implement both
            # of these. If passed a string, it's a column lookup, int is row
            # lookup. Some other tools do this. However, if that's the solution,
            # then our "ref extra is list of string" solution doesn't work.
            # TODO: fix
            or node.from_op.name.endswith("pick")
            or node.from_op.name.endswith("index")
        ):
            return None
        getitem_arg1 = list(node.from_op.inputs.values())[1]
        if not isinstance(getitem_arg1, graph.ConstNode):
            return None
        ref.extra.append(str(getitem_arg1.val))
    return ref


def ref_to_node(ref: ref_base.Ref) -> typing.Optional[graph.Node]:
    """Inverse of node_to_ref, see docstring for node_to_ref."""

    extra = ref.extra or []
    ref = copy.copy(ref)
    ref.extra = []

    from weave.ops_primitives import get

    node = get(str(ref))
    for str_key in extra:
        key: typing.Union[str, int] = str_key
        try:
            key = int(str_key)
        except ValueError:
            pass
        if hasattr(node, "__getitem__"):
            node = node[key]  # type: ignore
        elif hasattr(node, "pick"):
            node = node.pick(key)  # type: ignore
        else:
            return None
    return node



# See mappers_python_def for actual implementation. Split into two
# to resolve circular dependency.
import typing

if typing.TYPE_CHECKING:
    from . import mappers
    from . import artifact_base
    from . import weave_types


def map_to_python(  # type: ignore[empty-body]
    type_: "weave_types.Type",
    artifact: "artifact_base.Artifact",
    path: list[str] = [],
    mapper_options: typing.Any = None,
) -> "mappers.Mapper":
    raise NotImplementedError


def map_from_python(  # type: ignore[empty-body]
    type_: "weave_types.Type",
    artifact: "artifact_base.Artifact",
    path: list[str] = [],
    mapper_options: typing.Any = None,
) -> "mappers.Mapper":
    raise NotImplementedError



import json
import os
import random
import string
import urllib
import typing

from IPython.display import display
from IPython.display import IFrame

from . import context
from . import graph
from . import panel
from . import node_ref
from . import weave_types as types
from . import weavejs_fixes
from . import storage
from . import util
from . import errors
from . import usage_analytics
from . import ref_base
from . import artifact_fs
from . import ops


def make_varname_for_type(t: types.Type):
    if isinstance(t, types.List) and isinstance(t.object_type, types.TypedDict):
        return "table"
    return t.name


def make_container(
    obj: typing.Union[panel.Panel, graph.Node], name: str
) -> panel.Panel:
    from weave.panels import Group

    if isinstance(obj, graph.Node):
        return Group(
            items={name: obj},
            layoutMode="flow",
            enableAddPanel=True,
            showExpressions=True,
        )
    else:
        return obj


def make_show_obj(obj: typing.Any) -> tuple[typing.Union[panel.Panel, graph.Node], str]:
    node: graph.Node
    if obj is None:
        return graph.VoidNode(), "panel0"
    elif isinstance(obj, panel.Panel):
        return obj, obj.id + "0"
    elif isinstance(obj, graph.Node):
        ref = node_ref.node_to_ref(obj)
        if ref is not None and isinstance(ref, artifact_fs.FilesystemArtifactRef):
            name = ref.name
        else:
            name = make_varname_for_type(obj.type)

        return obj, name
    elif isinstance(obj, ref_base.Ref):
        if isinstance(obj, artifact_fs.FilesystemArtifactRef):
            name = obj.name
            uri = obj.branch_uri
        else:
            name = make_varname_for_type(obj.type)
            uri = obj.uri
        return ops.get(uri), name  # type: ignore

    if types.TypeRegistry.has_type(obj):
        names = util.find_names(obj)

        ref = storage.save(obj, name=names[-1])
        node = ops.get(ref.uri)  # type: ignore
        return node, make_varname_for_type(ref.type)

    raise errors.WeaveTypeError(
        "%s not yet supported. Create a weave.Type to add support." % type(obj)
    )


# Broken out into to separate function for testing
def _show_params(obj):
    # Get a weave node (expression) that will return the object, and something to call it
    node, var_name = make_show_obj(obj)

    # Place the node inside a root group, so the user can add panels.
    ui_root = make_container(node, var_name)

    # Name the panel artifact after the object's variable name for now,
    # and only create a new one if the name is not take yet.
    # If the name exists, we'll render whatever is already stored there!

    # Actually, disabled this... always create a new panel for now.

    panel_name = "dashboard-" + var_name
    # panel_ref = storage.get_local_version_ref(panel_name, "latest")
    panel_ref = storage.save(ui_root, name=f"{panel_name}:latest")

    # convert panel_ref to a a get expression.
    show_node = ops.get(panel_ref.branch_uri)

    # This fixes an issue with giant types causing the cell to take forever
    # for initial render (at least when using the dev server). The UI
    # will refine the type anyway.
    show_node.type = types.Any()

    return {"weave_node": weavejs_fixes.fixup_node(show_node)}


def show_url(obj=None):
    if os.environ.get("WEAVE_FRONTEND_DEVMODE"):
        context.use_frontend_devmode()
    params = _show_params(obj)
    panel_url = f"{context.get_frontend_url()}?fullScreen"
    if "weave_node" in params:
        panel_url += "&expNode=%s" % urllib.parse.quote(
            json.dumps(params["weave_node"].to_json())
        )
    if "panel_id" in params:
        panel_url += "&panelId=%s" % urllib.parse.quote(params["panel_id"])
    if "panel_config" in params:
        panel_url += "&panelConfig=%s" % urllib.parse.quote(
            json.dumps(params["panel_config"])
        )
    return panel_url


def show(obj: typing.Any = None, height: int = 400) -> typing.Any:
    if not util.is_notebook():
        usage_analytics.show_called({"error": True})
        raise RuntimeError(
            "`weave.show()` can only be called within notebooks. To extract the value of "
            "a weave node, try `weave.use()`."
        )

    if util.is_pandas_dataframe(obj):
        obj = ops.dataframe_to_arrow(obj)

    panel_url = show_url(obj)

    if util.is_colab():
        from google.colab.output import serve_kernel_port_as_iframe

        url = urllib.parse.urlparse(panel_url)
        usage_analytics.show_called({"colab": True})
        serve_kernel_port_as_iframe(url.port, path=url.path + "?" + url.query)
    else:
        usage_analytics.show_called()
        iframe = IFrame(panel_url, "100%", "%spx" % height, ['allow="clipboard-write"'])
        display(iframe)


def _ipython_display_method_(self):
    show(self)


# Inject _ipython_display_ methods on classes we want to automatically
# show when the last expression in a notebook cell produces them.
graph.Node._ipython_display_ = _ipython_display_method_  # type: ignore
# Don't automatically show weave panel in notebook for ref
# ref_base.Ref._ipython_display_ = _ipython_display_method_  # type: ignore



# This file contains helper functions to inspect nodes. The main
# purpose it to be used in debugging sessions

"""
NODE ID    NAME                                            PARAM        NODE TYPE          TYPE NAME          TYPE ID
---------  ----------------------------------------------  -----------  -----------------  ---------------  ---------
0          ** CONST_VAL                                                 ConstNode          Function                 0
1          f └- dict                                                    RuntimeOutputNode  TypedDict                7
2          f   ├- number-bin                               bin          OutputNode         TV(NoneType)             8
3          f   │ ├- pick                                   in           OutputNode         TV(NoneType)             8
4          f   │ │ ├- row                                  obj          VarNode            TV(TypedDict)            1
5          f   │ │ └- <<_step>>                            key          ConstNode          String                   3
6          f   │ └- numbers-binsequal                      binFn        OutputNode         Function                11
7          f   │   ├- list                                 arr          OutputNode         List                    14
8          f   │   │ ├- numbers-min                        a            OutputNode         TV(Number)              15
9          f   │   │ │ └- pick                             arr          OutputNode         TV(List)                17
10         f   │   │ │   ├- concat                         obj          OutputNode         TV(List)                21
11         f   │   │ │   │ └- run-history3                 arr          OutputNode         TV(List)                24
12         f   │   │ │   │   └- project-filteredRuns       run          OutputNode         TV(List)                28
13         f   │   │ │   │     ├- root-project             project      OutputNode         TV(projectType)         30
14         f   │   │ │   │     │ ├- <<shawn>>              entityName   ConstNode          String                   3
15         f   │   │ │   │     │ └- <<fasion-sweep>>       projectName  ConstNode          String                   3
16         f   │   │ │   │     ├- <<{"name":...":null}}>>  runFilters   ConstNode          String                   3
17         f   │   │ │   │     └- <<-createdAt>>           order        ConstNode          String                   3
.          f   │   │ │   └- NODE_ID(5)                     key          ConstNode          String                   3
18         f   │   │ └- numbers-max                        b            OutputNode         TV(Number)              15
.          f   │   │   └- NODE_ID(9)                       arr          OutputNode         TV(List)                17
19         f   │   └- <<50>>                               bins         ConstNode          Number                  12
20         f   └- run-name                                 run_name     OutputNode         TV(String)              10
21         f     └- tag-run                                run          OutputNode         TV(runType)             32
.          f       └- NODE_ID(4)                           obj          VarNode            TV(TypedDict)            1


**Types Table**
TYPE ID    TYPE NAME
---------  -----------------------------------
0          ** Function
1            ├- [input.row]: TV(TypedDict)
2            │ ├- [tag]: TypedDict
3            │ │ ├- [entityName]: String
.            │ │ ├- [projectName]: String
4            │ │ ├- [project]: projectType
.            │ │ ├- [runFilters]: String
.            │ │ ├- [order]: String
5            │ │ └- [run]: runType
6            │ └- [value]: TypedDict
7            └- [output_type]: TypedDict
8              ├- [bin]: TV(NoneType)
.              │ ├- [tag]: TYPE_ID(2)
9              │ └- [value]: NoneType
10             └- [run_name]: TV(String)
.                ├- [tag]: TYPE_ID(2)
.                └- [value]: String
11         ** Function
12           ├- [input.row]: Number
13           └- [output_type]: TypedDict
.              ├- [start]: Number
.              └- [stop]: Number
14         ** List
15           └- [object_type]: TV(Number)
16             ├- [tag]: TypedDict
.              │ ├- [entityName]: String
.              │ ├- [projectName]: String
.              │ ├- [project]: projectType
.              │ ├- [runFilters]: String
.              │ └- [order]: String
.              └- [value]: Number
17         ** TV(List)
.            ├- [tag]: TYPE_ID(16)
18           └- [value]: List
19             └- [object_type]: TV(NoneType)
20               ├- [tag]: TypedDict
.                │ └- [run]: runType
.                └- [value]: NoneType
21         ** TV(List)
.            ├- [tag]: TYPE_ID(16)
22           └- [value]: List
23             └- [object_type]: TV(TypedDict)
.                ├- [tag]: TYPE_ID(20)
.                └- [value]: TypedDict
24         ** TV(List)
.            ├- [tag]: TYPE_ID(16)
25           └- [value]: List
26             └- [object_type]: TV(List)
.                ├- [tag]: TYPE_ID(20)
27               └- [value]: List
.                  └- [object_type]: TypedDict
28         ** TV(List)
.            ├- [tag]: TYPE_ID(16)
29           └- [value]: List
.              └- [object_type]: runType
30         ** TV(projectType)
31           ├- [tag]: TypedDict
.            │ ├- [entityName]: String
.            │ └- [projectName]: String
.            └- [value]: projectType
32         ** TV(runType)
.            ├- [tag]: TYPE_ID(16)
.            └- [value]: runType

"""

import dataclasses
import json
import typing

import tabulate


from . import weave_types as types
from . import graph

from .partial_object import PartialObjectType


def _trimmed_string(s: str, max_len: int = 20) -> str:
    if len(s) > max_len - 3:
        return s[: max_len // 2] + "..." + s[-max_len // 2 :]
    return s


def _node_name(node: graph.Node) -> str:
    if isinstance(node, graph.OutputNode):
        return node.from_op.name
    elif isinstance(node, graph.VarNode):
        return node.name
    elif isinstance(node, graph.ConstNode):
        if isinstance(node.val, (str, int, float, bool)):
            return _trimmed_string("<<" + str(node.val) + ">>")
        return "CONST_VAL"
    elif isinstance(node, graph.VoidNode):
        return "VOID"
    else:
        raise Exception(f"Unknown node type: {type(node)}")


def _node_type_name(node: graph.Node) -> str:
    return type(node).__name__


def _type_name(node_type: types.Type) -> str:
    from .language_features.tagging import tagged_value_type

    if isinstance(node_type, tagged_value_type.TaggedValueType):
        return f"TV({_type_name(node_type.value)})"
    return str(type(node_type).__name__)


def _type_props(node_type: types.Type) -> typing.Dict[str, types.Type]:
    if isinstance(node_type, types.UnionType):
        return {str(n): m for n, m in enumerate(node_type.members)}
    elif isinstance(node_type, types.TypedDict):
        return node_type.property_types
    elif isinstance(node_type, types.Function):
        props = {f"input.{k}": v for k, v in node_type.input_types.items()}
        props["output_type"] = node_type.output_type
        return props
    elif isinstance(node_type, PartialObjectType):
        return node_type.keys
    return node_type.type_vars


@dataclasses.dataclass
class NodeIter:
    param_name: str
    node: graph.Node
    depth: int
    is_last: bool
    parent_node: typing.Optional[graph.Node] = None


@dataclasses.dataclass
class TypeIter:
    param_name: str
    node_type: types.Type
    depth: int
    is_last: bool
    parent_type: typing.Optional[types.Type] = None


def _type_iter(node_type: types.Type) -> typing.Iterator[TypeIter]:
    stack = [TypeIter("", node_type, 0, True)]
    while stack:
        type_iter = stack.pop(0)
        yield type_iter
        props = _type_props(type_iter.node_type)
        num_props = len(props)

        for ndx, (prop_name, prop_type) in reversed(list(enumerate(props.items()))):
            stack.insert(
                0,
                TypeIter(
                    prop_name,
                    prop_type,
                    type_iter.depth + 1,
                    is_last=num_props == ndx + 1,
                    parent_type=type_iter.node_type,
                ),
            )


def _print_type_table(
    type_iterator: typing.Iterator[TypeIter],
    type_to_id_map: dict[types.Type, int],
) -> None:
    table = []
    completed_nodes = set()
    reference_nodes = set()
    depth_state = []
    last_depth = -1
    for type_iter in type_iterator:
        if type_iter.depth > last_depth:
            depth_state.append("OPEN")
        last_depth = type_iter.depth
        depth_state = depth_state[: type_iter.depth + 1]
        if type_iter.is_last:
            depth_state[type_iter.depth] = "CLOSED"
            depth_state = depth_state[: type_iter.depth + 1]
        name_prefix = ""
        for depth, state in enumerate(depth_state):
            if state == "OPEN":
                if depth == type_iter.depth:
                    name_prefix += "├-"
                else:
                    name_prefix += "│ "
            elif state == "CLOSED":
                if depth == type_iter.depth:
                    if depth == 0:
                        name_prefix += "**"
                    else:
                        name_prefix += "└-"
                else:
                    name_prefix += "  "

        if type_iter.param_name:
            name_prefix += f" [{type_iter.param_name}]: "
        else:
            name_prefix += " "

        row = {
            "TYPE ID": type_to_id_map[type_iter.node_type],
            "TYPE NAME": name_prefix + _type_name(type_iter.node_type),
        }

        if type_iter.node_type in completed_nodes:
            reference_nodes.add(type_iter.node_type)
            if type_iter.depth > 0 and not (type_iter.parent_type in reference_nodes):
                row["TYPE ID"] = "."
                if len(_type_props(type_iter.node_type)) == 0:
                    row["TYPE NAME"] = name_prefix + _type_name(type_iter.node_type)
                else:
                    row["TYPE NAME"] = (
                        name_prefix + f"TYPE_ID({type_to_id_map[type_iter.node_type]})"
                    )
                table.append(row)
        else:
            row["TYPE ID"] = type_to_id_map[type_iter.node_type]
            row["TYPE NAME"] = name_prefix + _type_name(type_iter.node_type)
            table.append(row)
            completed_nodes.add(type_iter.node_type)

    print("\n**Types Table**")
    tabulate.PRESERVE_WHITESPACE = True
    print(tabulate.tabulate(table, headers="keys"))
    print("")


@dataclasses.dataclass
class TypeInspector:
    base_type: types.Type

    id_to_type_map: dict[int, types.Type] = dataclasses.field(default_factory=dict)
    type_to_id_map: dict[types.Type, int] = dataclasses.field(default_factory=dict)

    def __post_init__(self) -> None:
        self._build_maps()

    def _build_maps(self) -> None:
        for type_iter in self._pre_order_type_iter():
            if type_iter.node_type not in self.type_to_id_map:
                type_id = len(self.id_to_type_map)
                self.id_to_type_map[type_id] = type_iter.node_type
                self.type_to_id_map[type_iter.node_type] = type_id

    def _pre_order_type_iter(self) -> typing.Iterator[TypeIter]:
        for type_iter in _type_iter(self.base_type):
            yield type_iter

    def lookup(self, type_id: int) -> types.Type:
        target_type = self.id_to_type_map[type_id]
        return target_type

    def summarize(self) -> None:
        print("\Type Summary:")
        print("\nType as string:", str(self.base_type))
        _print_type_table(self._pre_order_type_iter(), self.type_to_id_map)


@dataclasses.dataclass
class NodeInspector:
    base_node: graph.Node

    id_to_node_map: dict[int, graph.Node] = dataclasses.field(default_factory=dict)
    node_to_id_map: dict[graph.Node, int] = dataclasses.field(default_factory=dict)

    id_to_type_map: dict[int, types.Type] = dataclasses.field(default_factory=dict)
    type_to_id_map: dict[types.Type, int] = dataclasses.field(default_factory=dict)

    def __post_init__(self) -> None:
        self._build_maps()

    def _pre_order_node_iter(self) -> typing.Iterator[NodeIter]:
        stack = [NodeIter("", self.base_node, 0, True)]
        while stack:
            node_iter = stack.pop(0)
            yield node_iter
            if isinstance(node_iter.node, graph.OutputNode):
                num_params = len(node_iter.node.from_op.inputs)
                items = enumerate(node_iter.node.from_op.inputs.items())
                reversed_items = reversed(list(items))
                for ndx, (param_name, param) in reversed_items:
                    stack.insert(
                        0,
                        NodeIter(
                            param_name,
                            param,
                            node_iter.depth + 1,
                            is_last=num_params == ndx + 1,
                            parent_node=node_iter.node,
                        ),
                    )
            elif isinstance(node_iter.node, graph.ConstNode) and isinstance(
                node_iter.node.val, graph.Node
            ):
                stack.insert(
                    0,
                    NodeIter(
                        node_iter.param_name,
                        node_iter.node.val,
                        node_iter.depth + 1,
                        is_last=True,
                        parent_node=node_iter.node,
                    ),
                )

    def _pre_order_type_iter(self) -> typing.Iterator[TypeIter]:
        for node_iter in self._pre_order_node_iter():
            for type_iter in _type_iter(node_iter.node.type):
                yield type_iter

    def _build_maps(self) -> None:
        for node_iter in self._pre_order_node_iter():
            if node_iter.node not in self.node_to_id_map:
                node_id = len(self.id_to_node_map)
                self.id_to_node_map[node_id] = node_iter.node
                self.node_to_id_map[node_iter.node] = node_id
            node_type = node_iter.node.type
            for type_iter in _type_iter(node_type):
                if type_iter.node_type not in self.type_to_id_map:
                    type_id = len(self.id_to_type_map)
                    self.id_to_type_map[type_id] = type_iter.node_type
                    self.type_to_id_map[type_iter.node_type] = type_id

    def lookup(self, node_id: int) -> graph.Node:
        target_node = self.id_to_node_map[node_id]
        return target_node

    def summarize(self) -> None:
        print("\nNode Summary:")
        print("\nNode as string:", self.base_node)
        self.print_node_table()
        print("\nType as string:", str(self.base_node.type))
        _print_type_table(self._pre_order_type_iter(), self.type_to_id_map)

    def print_node_table(self) -> None:
        table = []
        completed_nodes = set()
        reference_nodes = set()
        depth_state = []
        last_depth = -1
        for node_iter in self._pre_order_node_iter():
            if node_iter.depth > last_depth:
                depth_state.append("OPEN")
            last_depth = node_iter.depth
            depth_state = depth_state[: node_iter.depth + 1]
            if node_iter.is_last:
                depth_state[node_iter.depth] = "CLOSED"
                depth_state = depth_state[: node_iter.depth + 1]
            if isinstance(node_iter.node, graph.ConstNode) and isinstance(
                node_iter.node.val, graph.Node
            ):
                depth_state[node_iter.depth] = "FUNCTION"
            name_prefix = ""
            for depth, state in enumerate(depth_state):
                if state == "OPEN":
                    if depth == node_iter.depth:
                        name_prefix += "├-"
                    else:
                        name_prefix += "│ "
                elif state == "FUNCTION":
                    if depth == node_iter.depth:
                        if depth == 0:
                            name_prefix += "**"
                        else:
                            name_prefix += "└-"
                    else:
                        name_prefix += "f "
                elif state == "CLOSED":
                    if depth == node_iter.depth:
                        if depth == 0:
                            name_prefix += "**"
                        else:
                            name_prefix += "└-"
                    else:
                        name_prefix += "  "

            name_prefix += " "

            row = {
                "NODE ID": None,
                "NAME": None,
                "PARAM": node_iter.param_name,
                "NODE TYPE": _node_type_name(node_iter.node),
                "TYPE NAME": _type_name(node_iter.node.type),
                "TYPE ID": self.type_to_id_map[node_iter.node.type],
            }
            if node_iter.node in completed_nodes:
                reference_nodes.add(node_iter.node)
                if not (node_iter.parent_node in reference_nodes):
                    row["NODE ID"] = "."
                    row["NAME"] = (
                        name_prefix + f"NODE_ID({self.node_to_id_map[node_iter.node]})"
                    )
                    table.append(row)
            else:
                row["NODE ID"] = self.node_to_id_map[node_iter.node]
                row["NAME"] = name_prefix + _node_name(node_iter.node)
                table.append(row)
                completed_nodes.add(node_iter.node)

        print("\n**Graph Table**")
        tabulate.PRESERVE_WHITESPACE = True
        print(tabulate.tabulate(table, headers="keys"))
        print("")



"""Functions for determining which op is being called."""
from dataclasses import dataclass
import functools
import logging
import typing
import json

from .language_features.tagging.is_tag_getter import is_tag_getter
from .language_features.tagging.tagged_value_type import TaggedValueType

from . import weave_types as types
from . import op_def
from . import op_args
from . import registry_mem
from . import graph
from . import errors
from . import pyfunc_type_util
from . import util
from . import memo


# I originally wrote this thinking that we could always choose the more specific
# op, defined as the one that's input types are strict subtypes of the other.
# But consider mapped_add(List[number], number) and
# arrow_mapped_add(ArrowList[number], union[number, ArrowList[number]])
# (this is how we've actually defined those currently)
# ArrowList[number] is a subtype of number, but number is a subtype of
# union[number, ArrowList[number]].
# arrow_mapped_add is more specific in the first arg, but less specific in the
# second!
# So for now, we only consider the first argument.
# TODO: Solve. I see three potential solutions
#   1. Make mapped ops automatically handle arrays in their second argument
#   2. Split arrow ops into two ops, for array/scalar in second arg
#   3. Figure out some clever way to resolve the above ambiguity here
def _op_args_is_subtype(lhs: op_args.OpArgs, rhs: op_args.OpArgs) -> bool:
    """Returns true if rhs is subtype of lhs"""
    if isinstance(lhs, op_args.OpNamedArgs) and isinstance(rhs, op_args.OpNamedArgs):
        # PR: fixing to allow different length var lists to distinguish
        # (supports get_with_default style op)
        if len(lhs.arg_types) > len(rhs.arg_types):
            return False
        for self_type, other_type in list(
            zip(lhs.initial_arg_types.values(), rhs.initial_arg_types.values())
        )[:1]:
            if not self_type.assign_type(other_type):
                return False
        return True
    elif isinstance(lhs, op_args.OpNamedArgs) and isinstance(rhs, op_args.OpVarArgs):
        return all(t.assign_type(rhs.arg_type) for t in lhs.initial_arg_types.values())
    elif isinstance(lhs, op_args.OpVarArgs) and isinstance(rhs, op_args.OpNamedArgs):
        return all(lhs.arg_type.assign_type(t) for t in rhs.initial_arg_types.values())
    elif isinstance(lhs, op_args.OpVarArgs) and isinstance(rhs, op_args.OpVarArgs):
        return lhs.arg_type.assign_type(rhs.arg_type)
    else:
        raise errors.WeaveInternalError("unknown op_args types: %s, %s" % (lhs, rhs))


def _nullability_ambiguity_resolution_rule(
    candidates: list[op_def.OpDef], first_arg_type: types.Type
) -> list[op_def.OpDef]:
    # nullability, if the first argument is None or List[None], then any
    # of the candidates can handle it. Choose the first one (choosing the
    # first ensures we choose a tag getter if there is one).
    if first_arg_type and (
        types.NoneType().assign_type(first_arg_type)
        or types.List(types.NoneType()).assign_type(first_arg_type)
    ):
        return [candidates[0]]
    return candidates


def _is_mapped_op(op: op_def.OpDef) -> bool:
    return op.derived_from is not None and op.derived_from.derived_ops["mapped"] == op
    # Leaving this here, we may want to include this condition in the future
    # or isinstance(op, op_def.AutoTagHandlingArrowOpDef)


def _mapped_ambiguity_resolution_rule(
    candidates: list[op_def.OpDef], first_arg_type: types.Type
) -> list[op_def.OpDef]:
    non_mapped_candidates = [
        candidate for candidate in candidates if not _is_mapped_op(candidate)
    ]
    if len(non_mapped_candidates) > 0:
        return non_mapped_candidates
    return candidates


def _tagged_ambiguity_resolution_rule(
    candidates: list[op_def.OpDef], first_arg_type: types.Type
) -> list[op_def.OpDef]:
    non_tagged_candidates = [
        candidate for candidate in candidates if not is_tag_getter(candidate)
    ]
    if len(non_tagged_candidates) > 0:
        return non_tagged_candidates
    return candidates


def _subtype_sorting_ambiguity_resolution_rule_cmp(
    a: op_def.OpDef, b: op_def.OpDef
) -> int:
    # TODO: make this less hacky
    # If we're mapping contains, don't do substring matching
    ambiguous_contains = ["mapped_string-contains", "contains"]
    if a.name in ambiguous_contains and b.name in ambiguous_contains:
        if a.name == "contains":
            return -1
        else:
            return 1
    b_is_subtype = _op_args_is_subtype(a.input_type, b.input_type)
    a_is_subtype = _op_args_is_subtype(b.input_type, a.input_type)
    if a_is_subtype and b_is_subtype:
        raise errors.WeaveDispatchError(
            "Ambiguous ops %s, %s. Ops' input types are equivalent" % (a.name, b.name)
        )
    if a_is_subtype:
        return -1
    if b_is_subtype:
        return 1
    raise errors.WeaveDispatchError(
        "Ambiguous ops %s, %s. Ops' input types first arguments must be subset in one direction or the other."
        % (a.name, b.name)
    )


def _subtype_sorting_ambiguity_resolution_rule(
    candidates: list[op_def.OpDef], first_arg_type: types.Type
) -> list[op_def.OpDef]:
    return sorted(
        candidates,
        key=functools.cmp_to_key(_subtype_sorting_ambiguity_resolution_rule_cmp),
    )


def _apply_ambiguity_rules(
    candidates: list[op_def.OpDef],
    first_arg_type: types.Type,
    rules: list[
        typing.Tuple[
            str, typing.Callable[[list[op_def.OpDef], types.Type], list[op_def.OpDef]]
        ]
    ],
) -> list[op_def.OpDef]:
    for rule_name, rule in rules:
        reduced_candidates = rule(candidates, first_arg_type)
        if len(reduced_candidates) < len(candidates):
            logging.debug(
                f"Dispatch Ambiguity Resolution - {rule_name} Rule reduced set from {len(candidates)} to {len(reduced_candidates)}"
            )
        candidates = reduced_candidates
        if len(candidates) == 1:
            return candidates
    return candidates


def _resolve_op_ambiguity(
    candidates: list[op_def.OpDef], first_arg_type: types.Type
) -> op_def.OpDef:
    if len(candidates) == 1:
        return candidates[0]
    final_candidates = _apply_ambiguity_rules(
        candidates,
        first_arg_type,
        [
            ("Prefer Non-Mapped", _mapped_ambiguity_resolution_rule),
            ("If Null Input, Choose First", _nullability_ambiguity_resolution_rule),
            ("Prefer Non-Tagged", _tagged_ambiguity_resolution_rule),
            (
                "Prefer Sub Type of Super Type",
                _subtype_sorting_ambiguity_resolution_rule,
            ),
        ],
    )

    return final_candidates[0]


def _get_ops_by_name(fq_op_name: str) -> list[op_def.OpDef]:
    """Returns a single op that matches the given name and raw inputs (inputs can be python objects or nodes)"""
    shared_name_ops: list[op_def.OpDef]

    if "://" in fq_op_name:
        # If the incoming op is a locally-defined op, then we are just going to look at the derived op space.
        # We don't need to search the whole registry since by definition it is user-defined
        op = registry_mem.memory_registry.get_op(fq_op_name)
        derived_ops = list(op.derived_ops.values())
        shared_name_ops = [op] + derived_ops
    # Else, we lookup all the ops with the same common name
    else:
        shared_name_ops = registry_mem.memory_registry.find_ops_by_common_name(
            op_def.common_name(fq_op_name)
        )
    return shared_name_ops


def _dispatch_first_arg_inner(name: str, first_arg: types.Type) -> list[op_def.OpDef]:
    ops_with_name = _get_ops_by_name(name)
    ops_with_name_and_arg = []
    for op in ops_with_name:
        if op.input_type.first_param_valid(first_arg):
            ops_with_name_and_arg.append(op)

    return ops_with_name_and_arg


@memo.memo
def _dispatch_first_arg_cached(name: str, first_arg: types.Type) -> list[op_def.OpDef]:
    return _dispatch_first_arg_inner(name, first_arg)


def _dispatch_first_arg(name: str, first_arg: types.Type) -> list[op_def.OpDef]:
    # We want to memoize on the first argument, so that when there are many ops
    # hanging off a single node, we don't redo the dispatch for each. If there is
    # a Const in the Type, we can't hash it (this should be uncommon in the first
    # argument)
    # TODO: track stats
    if not isinstance(first_arg, types.Const):
        try:
            return _dispatch_first_arg_cached(name, first_arg)
        except errors.WeaveHashConstTypeError:
            pass
    return _dispatch_first_arg_inner(name, first_arg)


def _dispatch_remaining_args(
    first_arg_ops: list[op_def.OpDef], kwargs: dict[str, types.Type]
) -> list[op_def.OpDef]:
    candidates: list[op_def.OpDef] = []
    for op in first_arg_ops:
        if op.input_type.nonfirst_params_valid(list(kwargs.values())):
            candidates.append(op)
    return candidates


def get_op_for_inputs(name: str, kwargs: dict[str, types.Type]) -> op_def.OpDef:
    if not kwargs:
        # zero argument ops
        ops = _get_ops_by_name(name)
        if not ops:
            err = errors.WeaveDispatchError(
                f'Cannot dispatch op "{name}"; no matching op found'
            )
            util.raise_exception_with_sentry_if_available(err, [name])
        elif len(ops) > 1:
            err = errors.WeaveDispatchError(
                f'Cannot dispatch zero-arg op "{name}"; multiple matching ops found'
            )
            util.raise_exception_with_sentry_if_available(err, [name])
        return ops[0]

    if (
        name.startswith("panel_table")
        or name.startswith("Query")
        or name.startswith("Facet")
        or name.startswith("panel_plot")
    ):
        # The types don't work for TableState for some reason. This is hacked elsewhere..
        # Hack it some more :(
        # TODO: remove.
        ops = _get_ops_by_name(name)

        # Take that type system!
        ops = [
            o
            for o in ops
            if name.split("-")[0].removeprefix("panel_").lower()
            in json.dumps(list(o.input_type.arg_types.values())[0].to_dict()).lower()  # type: ignore
        ]

        return ops[0]

    input_keys = list(kwargs.keys())
    input_types = list(kwargs.values())

    # Dispatch first arg first. This is important for performance for two reasons:
    # 1. We memoize dispatch_first_arg, so we don't do duplicate work in the case
    #    where executing a graph with lots of fanout. This is common in Weave
    #    (e.g. some_list[0], some_list[1], some_list[2], ...)
    # 2. We don't have to check remaining types for ops that don't match the first
    #    type.
    ops = _dispatch_first_arg(name, input_types[0])
    if not ops:
        logging.info('No ops found for "%s" with first arg "%s"', name, input_types[0])
        err = errors.WeaveDispatchError(
            f'Cannot dispatch op "{name}"; no matching op found for first arg type: {input_types[0]}'
        )
        util.raise_exception_with_sentry_if_available(err, [name])

    final_ops = _dispatch_remaining_args(
        ops, dict(zip(input_keys[1:], input_types[1:]))
    )
    if not final_ops:
        err = errors.WeaveDispatchError(
            f'Cannot dispatch op "{name}"; ops {ops} matched first arg type, but no matching ops found for remaining arg types: {input_types[1:]}'
        )
        util.raise_exception_with_sentry_if_available(err, [name])

    return _resolve_op_ambiguity(final_ops, input_types[0])


def _type_of_input_param(v: typing.Any) -> types.Type:
    if isinstance(v, graph.Node):
        # Check if its a Node first, sometimes we mixin a callables with Node!
        if isinstance(v, graph.ConstNode) and not isinstance(v.type, types.Const):
            return types.Const(v.type, v.val)
        return v.type
    elif callable(v):
        input_type = pyfunc_type_util.determine_input_type(v, None, True)
        output_type = pyfunc_type_util.determine_output_type(v, None, True)
        if not isinstance(input_type, op_args.OpNamedArgs):
            raise errors.WeaveInternalError("Function conversion requires named args")
        if callable(output_type):
            raise errors.WeaveInternalError(
                "Function conversion does not support callable output types"
            )
        return types.Function(
            input_type.arg_types,
            output_type,
        )
    else:
        return types.Const(types.TypeRegistry.type_of(v), v)


def _get_self_bound_input_types(
    self_node: graph.Node, *args: list[typing.Any], **kwargs: dict[str, typing.Any]
) -> dict[str, types.Type]:
    input_types = {"self": _type_of_input_param(self_node)}
    input_types.update({str(i): _type_of_input_param(v) for i, v in enumerate(args)})
    input_types.update({k: _type_of_input_param(v) for k, v in kwargs.items()})
    return input_types


@dataclass
class BoundPotentialOpDefs:
    self_node: graph.Node
    potential_ops: list[op_def.OpDef]

    def __call__(self, *args: typing.Any, **kwargs: typing.Any) -> "RuntimeOutputNode":
        inputs = _get_self_bound_input_types(self.self_node, *args, **kwargs)
        input_keys = list(inputs.keys())
        input_types = list(inputs.values())
        ops = _dispatch_remaining_args(
            self.potential_ops, dict(zip(input_keys[1:], input_types[1:]))
        )
        if not ops:
            if any(callable(v) for v in args + tuple(kwargs.values())):
                # _dispatch_remaining_args is broken for callable args. They
                # are dispatchable, and this can be fixed! bind_params in op_def
                # does it correctly. "test_dispatch_lambda" tests this case.
                # TODO: add proper fix
                ops = self.potential_ops
            else:
                err = errors.WeaveDispatchError(
                    f'Cannot dispatch choose op from "{self.potential_ops}"; no matching op found'
                )
                util.raise_exception_with_sentry_if_available(
                    err, [str(self.potential_ops)]
                )
        op = _resolve_op_ambiguity(ops, input_types[0])
        params = op.input_type.create_param_dict([self.self_node] + list(args), kwargs)
        return op(**params)


def _dispatch_dunder(
    name: str,
) -> typing.Callable[..., "RuntimeOutputNode"]:
    def dispatch_dunder_inner(
        self_node: graph.Node, *args: typing.Any, **kwargs: typing.Any
    ) -> "RuntimeOutputNode":
        input_types = _get_self_bound_input_types(self_node, *args, **kwargs)
        op = get_op_for_inputs(name, input_types)
        params = op.input_type.create_param_dict([self_node] + list(args), kwargs)
        return op(**params)

    return dispatch_dunder_inner


class DispatchMixin:
    # Little hack, storage._get_ref expects to be able to check whether
    # any object hasattr('_ref') including nodes. Set it here so that
    # our __getattr__ op method doesn't handle that check.
    _ref = None
    # ipython tries to figure out if we have implemented a __getattr__
    # by checking for this attribute. But the weave.op() decorator makes
    # __getattr__ behave oddly, its now a lazy getattr that will always return
    # something. So add the attribute here to tell ipython that yes we do
    # have a __getattr__. This fixes Node._ipython_display()_ not getting fired.
    _ipython_canary_method_should_not_exist_ = None

    # Needed for storage.to_python hacks. Remove after those hacks are fixed.
    # TODO: fix
    to_pylist_notags = None
    as_py = None

    # This populates jupyter auto-complete, but it also really screws up
    # the vscode debugger
    # def __dir__(self) -> list[str]:
    #     ops = registry_mem.memory_registry.find_chainable_ops(self.type)
    #     return [o.common_name for o in ops]

    def __getattr__(self, attr: str) -> typing.Any:
        node_self = typing.cast(graph.Node, self)
        if attr.startswith("__") and attr.endswith("__"):
            return getattr(super(), attr)
        possible_bound_op = self._get_op(attr)
        if possible_bound_op is not None:
            return possible_bound_op
        possible_prop_node = self._get_prop(attr)
        if possible_prop_node is not None:
            return possible_prop_node
        raise errors.WeaveDispatchError(
            'No ops called "%s" available on type "%s"' % (attr, node_self.type)
        )

    # This implementation is not directly inside of __getattr__ so that
    # tests can call it directly in cases where attributes are overrode
    # by the node class (ex. name, type)
    def _get_op(
        self, attr: str
    ) -> typing.Optional[typing.Union[BoundPotentialOpDefs, op_def.BoundOpDef]]:
        node_self = typing.cast(graph.Node, self)
        ops_with_name_and_arg = _dispatch_first_arg(attr, node_self.type)
        if len(ops_with_name_and_arg) > 0:
            if len(ops_with_name_and_arg) == 1:
                # If there's only one candidate, we can just return it.
                return op_def.BoundOpDef(node_self, ops_with_name_and_arg[0])
            else:
                # Otherwise, we need to wait til we know the rest of the args
                # before we can decide which op to use.
                return BoundPotentialOpDefs(node_self, ops_with_name_and_arg)
        return None

    def _get_prop(self, attr: str) -> typing.Optional[graph.OutputNode]:
        node_self = typing.cast(graph.Node, self)
        use_type = types.non_none(node_self.type)
        if isinstance(use_type, TaggedValueType):
            use_type = use_type.value
        if isinstance(use_type, types.Function):
            use_type = use_type.output_type
        if isinstance(use_type, types.ObjectType):
            # Definitely an ObjectType
            if attr in use_type.property_types():
                obj_getattr = registry_mem.memory_registry.get_op("Object-__getattr__")
                return obj_getattr(node_self, attr)
            else:
                raise errors.WeaveDispatchError(
                    'No attributes called "%s" available on Object "%s"'
                    % (attr, node_self.type)
                )
        return None

    __call__ = _dispatch_dunder("__call__")
    __getitem__ = _dispatch_dunder("__getitem__")
    __len__ = _dispatch_dunder("__len__")
    __add__ = _dispatch_dunder("__add__")
    __sub__ = _dispatch_dunder("__sub__")
    __mul__ = _dispatch_dunder("__mul__")
    __truediv__ = _dispatch_dunder("__truediv__")
    __floordiv__ = _dispatch_dunder("__floordiv__")
    __pow__ = _dispatch_dunder("__pow__")
    __mod__ = _dispatch_dunder("__mod__")
    __round__ = _dispatch_dunder("__round__")
    __ge__ = _dispatch_dunder("__ge__")
    __gt__ = _dispatch_dunder("__gt__")
    __le__ = _dispatch_dunder("__le__")
    __lt__ = _dispatch_dunder("__lt__")
    __eq__ = _dispatch_dunder("__eq__")  # type: ignore
    __ne__ = _dispatch_dunder("__ne__")  # type: ignore
    __neg__ = _dispatch_dunder("__neg__")
    __contains__ = _dispatch_dunder("__contains__")
    __and__ = _dispatch_dunder("and")
    __or__ = _dispatch_dunder("and")


class RuntimeOutputNode(graph.OutputNode, DispatchMixin):
    pass


class RuntimeVarNode(graph.VarNode, DispatchMixin):
    pass


class RuntimeConstNode(graph.ConstNode, DispatchMixin):
    pass



import dataclasses
import typing

import weave
from .. import panel
from .. import weave_internal
from .. import graph


@weave.type()
class DropdownConfig:
    choices: weave.Node[list[str]] = dataclasses.field(
        default_factory=lambda: graph.VoidNode()
    )


@weave.type()
class Dropdown(panel.Panel):
    id = "Dropdown"
    config: typing.Optional[DropdownConfig] = dataclasses.field(
        default_factory=DropdownConfig
    )

    def __init__(
        self, input_node=graph.VoidNode(), vars=None, config=None, **options
    ) -> None:
        super().__init__(input_node=input_node, vars=vars)
        self.config = config
        if self.config is None:
            self.config = DropdownConfig()
        if "choices" in options:
            self.config.choices = options["choices"]
        if isinstance(self.input_node, weave.graph.VoidNode):
            choices_type = typing.cast(weave.types.List, self.config.choices.type)
            self.input_node = weave_internal.const(
                [], weave.types.List(choices_type.object_type)
            )

    @weave.op()
    def value(self) -> float:
        return weave.use(self.input_node)



import weave
import json
import typing

from ... import ops_arrow
from ... import op_def
from ... import compile

from . import gql_artifact_dag


@weave.type()
class RunChainSegmentInfo:
    run_name: str
    final_step: typing.Optional[int]


@weave.type()
class RunChain:
    entity_name: str
    project_name: str
    segments: list[RunChainSegmentInfo]

    def _history_node(self):
        with op_def.no_refine():
            proj = weave.ops.project(self.entity_name, self.project_name)
            history_nodes = []
            for seg in self.segments:
                # This is the only use of the deprecated history2 op in the Weave
                # codebase. We should be able to update to history3, but we'll
                # wait til we come back to working on this.
                hist_node = proj.run(seg.run_name).history2()
                if seg.final_step != None:
                    hist_node = hist_node.limit(seg.final_step)

                history_nodes.append(hist_node)

            history_node = weave.ops.List.concat(
                weave.ops.make_list(
                    **{f"node{i}": n for i, n in enumerate(history_nodes)}
                )
            )
            return history_node

    @weave.op()
    def refine_history_type(self) -> weave.types.Type:
        return compile.compile([self._history_node()])[0].type

    @weave.op(
        output_type=ops_arrow.ArrowWeaveListType(weave.types.TypedDict({})),
        refine_output_type=refine_history_type,
    )
    def history(self):
        return self._history_node()


# This is cacheable, since runs set their checkpoint artifact on startup.
@weave.op(render_info={"type": "function"})
def run_chain(run_path: str) -> RunChain:
    with op_def.no_refine():
        entity, project, run_id = run_path.split("/")
        proj = weave.ops.project(entity, project)
        run = proj.run(run_id)

        # Get the checkpoint artifact for this one if one exists
        used_arts = run.usedArtifactVersions()
        used_art_names = used_arts.name()
        used_art_type_names = used_arts.artifactType().name()
        with compile.enable_compile():
            res = weave.use((used_art_names, used_art_type_names))

        checkpoint_names = []
        for name, type_name in zip(res[0], res[1]):
            if type_name == "checkpoint":
                checkpoint_names.append(name)
        if len(checkpoint_names) == 0:
            return RunChain(
                entity_name=entity,
                project_name=project,
                segments=[RunChainSegmentInfo(run_name=run_id, final_step=None)],
            )
        if len(checkpoint_names) > 1:
            raise Exception("Multiple checkpoints found")
        checkpoint_name = checkpoint_names[0]

        # Get the sequence of runs and checkpoints that lead here.
        seq = gql_artifact_dag.get_run_checkpoint_chain(
            entity, project, checkpoint_name
        )

        # Fetch the contents of the checkpoint file for all checkpoints so we can get
        # the step for each.
        checkpoint_artifact_info = [s[1] for s in seq]
        nodes = []
        for artifact_info in checkpoint_artifact_info:
            checkpoint_json_node = (
                proj.artifactVersion(
                    artifact_info["artifactSequenceName"],
                    artifact_info["artifactCommitHash"],
                )
                .file("checkpoint.json")
                .contents()
            )
            nodes.append(checkpoint_json_node)

        with compile.enable_compile():
            checkpoint_jsons = weave.use(nodes)
        checkpoint_steps = [json.loads(c)["step"] for c in checkpoint_jsons]
        run_infos = [s[0] for s in seq]

        # Construct our segment objects.
        segments = []
        for run_info, checkpoint_step in zip(run_infos, checkpoint_steps):
            segments.append(
                RunChainSegmentInfo(
                    run_name=run_info["runName"], final_step=checkpoint_step
                )
            )
        segments.append(
            RunChainSegmentInfo(run_name=run_infos[-1]["runName"], final_step=None)
        )
        return RunChain(entity_name=entity, project_name=project, segments=segments)



import dataclasses
import typing

import weave
from .. import panel
from .. import panel_util
from .. import graph

ItemType = typing.TypeVar("ItemType")


@weave.type()
class LabeledItemConfig(typing.Generic[ItemType]):
    label: str = dataclasses.field(default_factory=lambda: "")
    item: ItemType = dataclasses.field(default_factory=lambda: graph.VoidNode())  # type: ignore


@weave.type()
class LabeledItem(panel.Panel):
    id = "LabeledItem"
    config: typing.Optional[LabeledItemConfig] = dataclasses.field(
        default_factory=lambda: None
    )

    def __init__(self, input_node=graph.VoidNode(), vars=None, config=None, **options):
        super().__init__(input_node=input_node, vars=vars)
        self.config = config
        if self.config is None:
            self.config = LabeledItemConfig(
                options["label"],
                panel_util.child_item(options["item"]),
            )



import contextlib
import dataclasses
import os
import random
import shutil
import pathlib
import typing

from . import uris
from . import errors
from . import wandb_client_api
from . import memo

from . import file_util
from . import environment as weave_env
from . import weave_types as types
from . import artifact_fs
from . import filesystem

if typing.TYPE_CHECKING:
    from wandb.apis.public import Run as WBRun


@memo.memo
def get_wandb_read_run(path: str) -> "WBRun":
    return wandb_client_api.wandb_public_api().run(path)  # type: ignore[no-untyped-call]


def wandb_run_dir() -> str:
    d = os.path.join(filesystem.get_filesystem_dir(), "_wandb_runs")
    os.makedirs(d, exist_ok=True)
    return d


@contextlib.contextmanager
def _isolated_download_and_atomic_mover(
    end_path: str,
) -> typing.Generator[typing.Tuple[str, typing.Callable[[str], None]], None, None]:
    rand_part = "".join(random.choice("0123456789ABCDEF") for _ in range(16))
    tmp_dir = os.path.join(wandb_run_dir(), f"tmp_{rand_part}")
    os.makedirs(tmp_dir, exist_ok=True)

    def mover(tmp_path: str) -> None:
        # This uses the same technique as WB artifacts
        pathlib.Path(end_path).parent.mkdir(parents=True, exist_ok=True)
        try:
            os.replace(tmp_path, end_path)
        except AttributeError:
            os.rename(tmp_path, end_path)

    try:
        yield tmp_dir, mover
    finally:
        shutil.rmtree(tmp_dir)


class WandbRunFiles(artifact_fs.FilesystemArtifact):
    def __init__(
        self,
        name: str,
        uri: "WeaveWBRunFilesURI",
    ):
        self.name = name
        self._run_files_uri = uri
        self._read_run = None
        self._local_path: dict[str, str] = {}
        self._ref = WandbRunFilesRef(self, None)

    @property
    def run(self) -> "WBRun":
        if self._read_run is None:
            self._read_run = get_wandb_read_run(self.name)
        if self._read_run is None:
            raise errors.WeaveInternalError(f"Could not read run {self.name}")
        return self._read_run

    @property
    def is_saved(self) -> bool:
        # Note: I think this is OK to always be true, since run files are
        # effectively always saved. But conceptually, the run could still be
        # "open" which might warrant returning `False`.
        return True

    @property
    def uri_obj(self) -> "WeaveWBRunFilesURI":
        return self._run_files_uri

    def direct_url(self, path: str) -> typing.Optional[str]:
        base_url = weave_env.wandb_base_url()
        uri = self.uri_obj
        return f"{base_url}/files/{uri.entity_name}/{uri.project_name}/{uri.run_name}/{path}"

    def path(self, path: str) -> str:
        # TODO: Move this logic to `io_service`. We can get rid of the
        # `get_wandb_read_run` as well once we do this

        # First, check if we already downloaded this file:
        if path in self._local_path:
            return self._local_path[path]

        static_file_path = os.path.join(wandb_run_dir(), str(self.name), path)
        # Next, check if another process has already downloaded this file:
        if os.path.exists(static_file_path):
            self._local_path[path] = static_file_path
            return static_file_path

        # Finally, download the file in an isolated directory:
        with _isolated_download_and_atomic_mover(static_file_path) as (tmp_dir, mover):
            with self.run.file(path).download(tmp_dir, replace=True) as fp:  # type: ignore[no-untyped-call]
                downloaded_file_path = fp.name
            mover(downloaded_file_path)

        self._local_path[path] = static_file_path
        return static_file_path

    def _path_info(self, path: str) -> "artifact_fs.FilesystemArtifactFile":
        # TODO: In the WBArtifact sister class, we check the manifest to see if
        # a) the file exists
        # b) it is a directory.
        #
        # Here, we blindly assume that the file exists and is not a directory.
        #
        # TODO: Enforce this check similarly.
        return artifact_fs.FilesystemArtifactFile(self, path)

    @contextlib.contextmanager
    def open(
        self, path: str, binary: bool = False
    ) -> typing.Generator[typing.IO, None, None]:
        mode = "rb" if binary else "r"
        p = self.path(path)
        with file_util.safe_open(p, mode) as f:
            yield f


class WandbRunFilesType(artifact_fs.FilesystemArtifactType):
    def save_instance(
        self, obj: typing.Any, artifact: typing.Any, name: str
    ) -> "WandbRunFilesRef":
        return WandbRunFilesRef(obj, None)


WandbRunFilesType.instance_classes = WandbRunFiles


class WandbRunFilesRef(artifact_fs.FilesystemArtifactRef):
    artifact: WandbRunFiles

    def versions(self) -> list[artifact_fs.FilesystemArtifactRef]:
        return [self]

    @classmethod
    def from_uri(cls, uri: "uris.WeaveURI") -> "WandbRunFilesRef":
        if not isinstance(uri, (WeaveWBRunFilesURI)):
            raise errors.WeaveInternalError(
                f"Invalid URI class passed to WandbRunFilesRef.from_uri: {type(uri)}"
            )
        return cls(
            WandbRunFiles(uri.name, uri=uri),
            path=uri.path,
        )


WandbRunFiles.RefClass = WandbRunFilesRef


@dataclasses.dataclass(frozen=True)
class WandbRunFilesRefType(types.RefType):
    pass


WandbRunFilesRefType.instance_class = WandbRunFilesRef
WandbRunFilesRefType.instance_classes = WandbRunFilesRef


@dataclasses.dataclass
class WeaveWBRunFilesURI(uris.WeaveURI):
    SCHEME = "wandb-run-file"
    entity_name: str
    project_name: str
    run_name: str
    path: typing.Optional[str] = None

    @classmethod
    def from_parsed_uri(
        cls,
        uri: str,
        schema: str,
        netloc: str,
        path: str,
        params: str,
        query: dict[str, list[str]],
        fragment: str,
    ) -> "WeaveWBRunFilesURI":
        path = path.strip("/")
        entity_name = netloc
        has_file_path = path.count("/") > 1
        if has_file_path:
            project_name, run_name, path = path.split("/", 2)
        else:
            project_name, run_name = path.split("/", 1)
            path = ""
        return cls(
            f"{entity_name}/{project_name}/{run_name}",
            None,
            entity_name,
            project_name,
            run_name,
            path or None,
        )

    @classmethod
    def from_run_identifiers(
        cls,
        entity_name: str,
        project_name: str,
        run_name: str,
    ) -> "WeaveWBRunFilesURI":
        return cls(
            f"{entity_name}/{project_name}/{run_name}",
            None,
            entity_name,
            project_name,
            run_name,
        )

    def to_ref(self) -> WandbRunFilesRef:
        return WandbRunFilesRef.from_uri(self)

    def __str__(self) -> str:
        netloc = f"{self.entity_name}/{self.project_name}/{self.run_name}"
        path = self.path or ""
        if path != "":
            path = f"/{path}"
        return f"{self.SCHEME}://{netloc}{path}"



import logging

from . import graph
from . import registry_mem
from . import weave_internal
from . import weave_types as types
from . import errors
from . import box
from . import compile
from . import engine_trace
from . import forward_graph
from . import language_nullability
from .language_features.tagging import tag_store

from . import ref_base
from . import op_policy

# from . import object_context


def _execute_fn_no_engine(item, index, map_fn):
    # executes map_fn without using the execute engine.
    if isinstance(map_fn, graph.OutputNode):
        inputs = {
            name: _execute_fn_no_engine(item, index, node)
            for name, node in map_fn.from_op.inputs.items()
        }
        op_def = registry_mem.memory_registry.get_op(map_fn.from_op.name)
        if language_nullability.should_force_none_result(inputs, op_def):
            return None
        for key in inputs:
            if isinstance(inputs[key], ref_base.Ref):
                inputs[key] = inputs[key].get()
        return op_def.resolve_fn(**inputs)
    elif isinstance(map_fn, graph.ConstNode):
        return map_fn.val
    elif isinstance(map_fn, graph.VarNode):
        if map_fn.name == "row":
            return item
        elif map_fn.name == "index":
            return index
        else:
            raise errors.WeaveInternalError(
                "Encountered unknown variable: %s" % map_fn.name
            )
    else:
        raise errors.WeaveInternalError("Invalid Node: %s" % map_fn)


def _resolve_static_branches(map_fn):
    result_store = forward_graph.get_node_result_store()
    if isinstance(map_fn, graph.OutputNode):
        if result_store.has(map_fn):
            return graph.ConstNode(map_fn.type, result_store[map_fn])

        inputs = {
            name: _resolve_static_branches(node)
            for name, node in map_fn.from_op.inputs.items()
        }

        if map_fn.from_op.name == "function-__call__":
            # Special case to expand function calls
            self = inputs["self"]
            if isinstance(self, graph.ConstNode):
                if isinstance(self.val, graph.ConstNode) and isinstance(
                    self.val.type, types.Function
                ):
                    self = self.val

                res = weave_internal.better_call_fn(
                    self,
                    inputs["arg1"],
                )
                fixed = compile.compile_fix_calls([res])[0]
                return _resolve_static_branches(fixed)

        if all(isinstance(v, graph.ConstNode) for v in inputs.values()):
            call_node = graph.OutputNode(map_fn.type, map_fn.from_op.name, inputs)

            # Don't compile here, we've already compiled.
            with compile.disable_compile():
                res = weave_internal.use(call_node)
            result_store[map_fn] = res
            return graph.ConstNode(map_fn.type, res)
        return graph.OutputNode(map_fn.type, map_fn.from_op.name, inputs)
    elif isinstance(map_fn, graph.ConstNode):
        return map_fn
    elif isinstance(map_fn, graph.VarNode):
        return map_fn
    else:
        raise errors.WeaveInternalError("Invalid Node: %s" % map_fn)


def op_can_be_async(op_name: str) -> bool:
    try:
        op = registry_mem.memory_registry.get_op(op_name)
    except errors.WeaveMissingOpDefError:
        return any(
            [
                o.is_async
                for o in registry_mem.memory_registry.find_ops_by_common_name(op_name)
            ]
        )
    return op.is_async


def _can_fast_map(map_fn):
    not_fastmappable_nodes = graph.filter_nodes_full(
        [map_fn],
        lambda n: isinstance(n, graph.OutputNode)
        and (
            op_can_be_async(n.from_op.name)
            or "://" in n.from_op.name
            or op_policy.should_cache(n.from_op.name)
            or op_policy.should_run_in_parallel(n.from_op.name)
        ),
    )
    return len(not_fastmappable_nodes) == 0


def _slow_map_fn(input_list, map_fn):
    calls = []
    for i, row in enumerate(input_list):
        calls.append(
            weave_internal.call_fn(
                map_fn,
                {
                    "row": graph.ConstNode(types.TypeRegistry.type_of(row), row),
                    "index": graph.ConstNode(types.Number(), i),
                },
            )
        )
    return weave_internal.use(calls)


def fast_map_fn(input_list, map_fn):
    """Maps a weave function over an input list, without using engine."""

    if len(input_list) == 0:
        return []

    # TODO: Perform this recursively in the main compile pass
    tracer = engine_trace.tracer()

    # we can resolve any branches that do not have variable node ancestors
    # one time up front.
    with tracer.trace("fast_map:resolve_static"):
        map_fn = _resolve_static_branches(map_fn)
    # Need to compile again after resolving static branches.

    if not _can_fast_map(map_fn):
        logging.warning("Cannot fast map, falling back to slow map for: %s", map_fn)
        return _slow_map_fn(input_list, map_fn)

    # For example, we may have fetched an expression out of a Const node.
    with tracer.trace("fast_map:compile2"):
        map_fn = compile.compile([map_fn])[0]

    logging.info("Fast mapping with map_fn: %s", map_fn)
    # now map the remaining weave_fn (after resolving static branches above)
    # over the input list
    with tracer.trace("fast_map:map"):
        list_tags = (
            None
            if not tag_store.is_tagged(input_list)
            else tag_store.get_tags(input_list)
        )
        result = []
        for i, item in enumerate(input_list):
            item = box.box(item)
            if list_tags is not None:
                # push down list tags to elements, mirroring arrow map (apply_fn_node_with_tag_pushdown)
                tag_store.add_tags(
                    item, list_tags, give_precedence_to_existing_tags=True
                )

            item_result = _execute_fn_no_engine(item, i, map_fn)
            result.append(item_result)
        return result



import contextlib
import typing

from . import context_state
from . import errors

if typing.TYPE_CHECKING:
    from .graph_client import GraphClient


@contextlib.contextmanager
def set_graph_client(
    client: typing.Optional["GraphClient"],
) -> typing.Iterator[typing.Optional["GraphClient"]]:
    client_token = context_state._graph_client.set(client)
    try:
        yield client
    finally:
        context_state._graph_client.reset(client_token)


def get_graph_client() -> typing.Optional["GraphClient"]:
    return context_state._graph_client.get()


def require_graph_client() -> "GraphClient":
    client = get_graph_client()
    if not client:
        raise errors.WeaveInitError("You must call `weave.init(<project_name>)` first")
    return client



import typing
from typing import Protocol, Generic, Sequence, TypeVar, Optional, Any
from abc import abstractmethod

from .ref_base import Ref
from .uris import WeaveURI
from .run import RunKey, Run

R = TypeVar("R", bound=Run)

if typing.TYPE_CHECKING:
    from .op_def import OpDef


class GraphClient(Protocol, Generic[R]):
    ##### Read API

    @abstractmethod
    def runs(self) -> Sequence[Run]:
        ...

    @abstractmethod
    def run(self, run_id: str) -> Optional[Run]:
        ...

    @abstractmethod
    def find_op_run(self, op_name: str, inputs: dict[str, Any]) -> Optional[Run]:
        ...

    @abstractmethod
    def run_children(self, run_id: str) -> Sequence[Run]:
        ...

    @abstractmethod
    def op_runs(self, op_def: "OpDef") -> Sequence[Run]:
        ...

    @abstractmethod
    def ref_input_to(self, ref: Ref) -> Sequence[Run]:
        ...

    @abstractmethod
    def ref_value_input_to(self, ref: Ref) -> Sequence[Run]:
        ...

    @abstractmethod
    def ref_output_of(self, ref: Ref) -> Optional[Run]:
        ...

    @abstractmethod
    def run_feedback(self, run_id: str) -> Sequence[dict[str, Any]]:
        ...

    @abstractmethod
    def feedback(self, feedback_id: str) -> Optional[dict[str, Any]]:
        ...

    ##### Helpers

    @abstractmethod
    def ref_is_own(self, ref: Optional[Ref]) -> bool:
        ...

    @abstractmethod
    def ref_uri(self, name: str, version: str, path: str) -> WeaveURI:
        ...

    @abstractmethod
    def run_ui_url(self, run: Run) -> str:
        ...

    ##### Write API

    @abstractmethod
    def save_object(self, obj: Any, name: str, branch_name: str) -> Ref:
        ...

    @abstractmethod
    def create_run(
        self,
        op_name: str,
        parent: Optional["RunKey"],
        inputs: dict[str, Any],
        input_refs: Sequence[Ref],
    ) -> R:
        ...

    @abstractmethod
    def fail_run(self, run: R, exception: BaseException) -> None:
        ...

    @abstractmethod
    def finish_run(self, run: R, output: Any, output_refs: Sequence[Ref]) -> None:
        ...

    @abstractmethod
    def add_feedback(self, run_id: str, feedback: dict[str, Any]) -> None:
        ...



import typing
import pyarrow as pa
import numpy as np

from ..arrow.arrow import arrow_as_array, offsets_starting_at_zero

from ..decorator_op import op
from ..op_args import OpVarArgs
from ..decorator_arrow_op import arrow_op
from .. import weave_types as types
from .. import _dict_utils
from ..ops_primitives import projection_utils
from ..language_features.tagging import (
    process_opdef_output_type,
)
from ..arrow.arrow_tags import direct_add_arrow_tags
from ..arrow import convert

from ..arrow.constructors import (
    vectorized_container_constructor_preprocessor,
    vectorized_input_types,
)
from ..arrow.list_ import ArrowWeaveList, ArrowWeaveListType


def typeddict_pick_output_type(input_types):
    output_type = _dict_utils.typeddict_pick_output_type(input_types)
    return process_opdef_output_type.op_make_type_tagged_resolver(
        output_type,
        process_opdef_output_type.op_get_tag_type_resolver(input_types["self"]),
    )


def _struct_array_field_names(array: pa.StructArray) -> list[str]:
    return [array.type.field(i).name for i in range(array.type.num_fields)]


def _arrow_val_tag_wrapper(
    array: pa.Array,
) -> typing.Tuple[typing.Callable[[pa.Array], pa.Array], pa.Array]:
    tags = None
    if isinstance(array, pa.StructArray):
        fields = _struct_array_field_names(array)
        if len(fields) == 2 and set(fields) == set(["_tag", "_value"]):
            tags = array.field("_tag")
            array = array.field("_value")

    def wrapper(res: pa.Array) -> typing.Any:
        if tags is not None:
            return direct_add_arrow_tags(res, tags)
        else:
            return res

    return wrapper, array


def _awl_pick(array: pa.Array, path: list[str]) -> pa.Array:
    def default_return():
        return pa.array([None] * len(array))

    if len(path) == 0:
        return default_return()
    tag_wrapper, inner_array = _arrow_val_tag_wrapper(array)
    key = path[0]
    next_path = path[1:]
    if isinstance(inner_array, pa.ListArray) and key == "*":
        if len(next_path) == 0:
            return array
        else:
            flattened_results = _awl_pick(inner_array.flatten(), next_path)
            rolled_results = pa.ListArray.from_arrays(
                offsets_starting_at_zero(inner_array), flattened_results
            )
            return tag_wrapper(rolled_results)
    elif isinstance(inner_array, pa.StructArray):
        all_names = _struct_array_field_names(inner_array)
        if key == "*":
            all_columns = [
                _awl_pick(inner_array.field(name), next_path) for name in all_names
            ]
            return tag_wrapper(pa.StructArray.from_arrays(all_columns, all_names))
        else:
            if key not in all_names:
                return default_return()
            key_val = inner_array.field(key)
            if len(next_path) == 0:
                res = key_val
            else:
                res = _awl_pick(key_val, next_path)
            return tag_wrapper(res)
    else:
        return default_return()


@arrow_op(
    name="ArrowWeaveListTypedDict-pick",
    input_type={"self": ArrowWeaveListType(types.TypedDict({})), "key": types.String()},
    output_type=lambda input_types: ArrowWeaveListType(
        typeddict_pick_output_type(
            {"self": input_types["self"].object_type, "key": input_types["key"]}
        )
    ),
    all_args_nullable=False,
)
def pick(self, key):
    object_type = typeddict_pick_output_type(
        {"self": self.object_type, "key": types.Const(types.String(), key)}
    )
    data = arrow_as_array(self._arrow_data)
    path = _dict_utils.split_escaped_string(key)
    result = _awl_pick(data, path)
    return ArrowWeaveList(result, object_type, self._artifact)


def _ensure_awl_for_merge(
    ensure_target: typing.Union[dict, ArrowWeaveList],
    compare_target: typing.Union[dict, ArrowWeaveList],
) -> ArrowWeaveList:
    if not isinstance(ensure_target, ArrowWeaveList):
        assert isinstance(compare_target, ArrowWeaveList)
        utility_awl = convert.to_arrow([ensure_target])
        return ArrowWeaveList(
            pa.repeat(ensure_target, len(compare_target)),
            utility_awl.object_type,
            utility_awl._artifact,
        )
    return ensure_target


MaybeTypedDictType = types.optional(types.TypedDict({}))
AWLMaybeTypedDictType = ArrowWeaveListType(MaybeTypedDictType)


@op(
    name="ArrowWeaveList-_preprocessMerge",
    input_type={
        "self": types.union(AWLMaybeTypedDictType, MaybeTypedDictType),
        "other": (
            lambda input_types: AWLMaybeTypedDictType
            if MaybeTypedDictType.assign_type(input_types["self"])
            else types.union(AWLMaybeTypedDictType, MaybeTypedDictType)
        ),
    },
    output_type=(
        lambda input_types: input_types["self"]
        if ArrowWeaveListType().assign_type(input_types["self"])
        else ArrowWeaveListType(input_types["self"])
    ),
    hidden=True,
)
def preprocess_merge(self, other):
    "Preprocesses the merge operation to ensure that both inputs are ArrowWeaveLists."
    return _ensure_awl_for_merge(self, other)


@arrow_op(
    name="ArrowWeaveListTypedDict-merge",
    input_type={
        # Optional inner type to support nullability
        "self": ArrowWeaveListType(types.optional(types.TypedDict({}))),
        "other": ArrowWeaveListType(types.TypedDict({})),
    },
    output_type=lambda input_types: ArrowWeaveListType(
        _dict_utils.typeddict_merge_output_type(
            {
                "self": input_types["self"].object_type,
                "other": input_types["other"].object_type,
            }
        )
    ),
    all_args_nullable=False,
)
def merge(self, other):
    field_arrays: dict[str, pa.Array] = {}
    for arrow_weave_list in (self, other):
        if isinstance(arrow_weave_list.object_type, types.NoneType):
            continue
        for key in arrow_weave_list.object_type.property_types:
            if isinstance(arrow_weave_list._arrow_data, pa.Table):
                field_arrays[key] = arrow_weave_list._arrow_data[key].combine_chunks()
            else:
                field_arrays[key] = arrow_weave_list._arrow_data.field(key)

    if isinstance(self._arrow_data, pa.Array):
        mask = pa.compute.is_null(self._arrow_data)
    else:
        # TODO: implement for table, chunkedarray etc
        mask = None

    field_names, arrays = tuple(zip(*field_arrays.items()))
    new_array = pa.StructArray.from_arrays(arrays=arrays, names=field_names, mask=mask)  # type: ignore

    result = ArrowWeaveList(
        new_array,
        _dict_utils.typeddict_merge_output_type(
            {"self": self.object_type, "other": other.object_type}
        ),
        self._artifact,
    )

    return result


# this function handles the following case:
# types.TypeRegistry.type_of(awl1) == tagged_value_type.TaggedValueType(
#    types.TypedDict({"outer1": types.String()}), ArrowWeaveListType(types.Int())
# )
# types.TypeRegistry.type_of(awl2) == tagged_value_type.TaggedValueType(
#    types.TypedDict({"outer2": types.String()}), ArrowWeaveListType(types.Int())
# )
#
# push down tags on list to tags on dict elements
# types.TypeRegistry.type_of(arrow_dict_(a=awl1, b=awl2)) == ArrowWeaveListType(
#    types.TypedDict(
#        {
#            "a": tagged_value_type.TaggedValueType(
#                types.TypedDict({"outer1": types.String()}), types.Int()
#            ),
#            "b": tagged_value_type.TaggedValueType(
#                types.TypedDict({"outer2": types.String()}), types.Int()
#            ),
#        }
#    )
# )


def vectorized_dict_output_type(input_types):
    prop_types = vectorized_input_types(input_types)
    return ArrowWeaveListType(types.TypedDict(prop_types))


@op(
    name="ArrowWeaveList-vectorizedDict",
    input_type=OpVarArgs(types.Any()),
    output_type=vectorized_dict_output_type,
    render_info={"type": "function"},
)
def arrow_dict_(**d):
    res = vectorized_container_constructor_preprocessor(d)
    table = pa.Table.from_arrays(res.arrays, list(d.keys()))
    return ArrowWeaveList(table, types.TypedDict(res.prop_types), res.artifact)


@arrow_op(
    name="ArrowWeaveList-2DProjection",
    input_type={
        "table": ArrowWeaveListType(types.TypedDict({})),
        "projectionAlgorithm": types.String(),
        "inputCardinality": types.String(),
        "inputColumnNames": types.List(types.String()),
        "algorithmOptions": types.TypedDict({}),
    },
    output_type=lambda input_types: ArrowWeaveListType(
        types.TypedDict(
            {
                "projection": types.TypedDict(
                    {"x": types.Number(), "y": types.Number()}
                ),
                "source": input_types["table"].object_type,
            }
        )
    ),
)
def awl_2d_projection(
    table,
    projectionAlgorithm,
    inputCardinality,
    inputColumnNames,
    algorithmOptions,
):
    inputColumnNames = list(set(inputColumnNames))
    source_column = arrow_as_array(table._arrow_data)
    data = table._arrow_data_asarray_no_tags()

    def default_projection():
        return np.zeros((len(data), 2))

    if len(inputColumnNames) == 0 or len(data) < 2:
        np_projection = default_projection()
    else:
        if inputCardinality == "single":
            path = _dict_utils.split_escaped_string(inputColumnNames[0])
            np_array_of_embeddings = np.stack(
                _awl_pick(data, path).to_numpy(False), axis=0
            )
        else:
            column_data = [
                _awl_pick(data, _dict_utils.split_escaped_string(c))
                for c in inputColumnNames
            ]
            # Filter out null columns
            column_data = [
                col for col in column_data if not isinstance(col, pa.NullArray)
            ]
            np_array_of_embeddings = np.array(column_data).T

        # coerce to float to turn Nones into nans
        np_array_of_embeddings = np_array_of_embeddings.astype("f")

        # impute nans with average value for column - this handles the case
        # where we have a column of Optional[number]
        none_indices = np.argwhere(np.isnan(np_array_of_embeddings))
        column_indices = none_indices[:, 1]

        np.put(
            np_array_of_embeddings,
            none_indices,
            np.nanmean(np_array_of_embeddings, axis=0)[column_indices],
        )

        # if any remaining nans (can only happen if all values in a column are nan), replace with zeros
        # such columns will be removed in the next step
        np_array_of_embeddings = np.nan_to_num(np_array_of_embeddings)

        # remove 0-only columns
        np_array_of_embeddings = np_array_of_embeddings[
            :, ~((np_array_of_embeddings == 0).all(axis=0))
        ]
        # If the selected data is not a 2D array, or if it has less than 2 columns, then
        # we can't perform a 2D projection. In this case, we just return a 2D array of
        # zeros.
        if (
            len(np_array_of_embeddings.shape) != 2
            or np_array_of_embeddings.shape[1] < 2
        ):
            np_projection = default_projection()
        else:
            np_array_of_embeddings = np.nan_to_num(np_array_of_embeddings, copy=False)
            np_projection = projection_utils.perform_2D_projection_with_timeout(
                np_array_of_embeddings, projectionAlgorithm, algorithmOptions
            )
    x_column = np_projection[:, 0]
    y_column = np_projection[:, 1]

    projection_column = pa.StructArray.from_arrays(
        [pa.array(x_column), pa.array(y_column)], ["x", "y"]
    )

    projection_column = pa.StructArray.from_arrays(
        [projection_column, source_column], ["projection", "source"]
    )

    return ArrowWeaveList(
        projection_column,
        types.TypedDict(
            {
                "projection": types.TypedDict(
                    {
                        "x": types.Number(),
                        "y": types.Number(),
                    }
                ),
                "source": table.object_type,
            }
        ),
        table._artifact,
    )


@arrow_op(
    name="ArrowWeaveList-projection2D",
    input_type={
        "table": ArrowWeaveListType(types.TypedDict({})),
        "projectionAlgorithm": types.String(),
        "inputCardinality": types.String(),
        "inputColumnNames": types.List(types.String()),
        "algorithmOptions": types.TypedDict({}),
    },
    output_type=lambda input_types: ArrowWeaveListType(
        types.TypedDict(
            {
                "projection": types.TypedDict(
                    {"x": types.Number(), "y": types.Number()}
                ),
                "source": input_types["table"].object_type,
            }
        )
    ),
)
def awl_projection_2d(
    table,
    projectionAlgorithm,
    inputCardinality,
    inputColumnNames,
    algorithmOptions,
):
    inputColumnNames = list(set(inputColumnNames))
    source_column = arrow_as_array(table._arrow_data)
    data = table._arrow_data_asarray_no_tags()

    def default_projection():
        return np.array([[0, 0] for row in data])

    if len(inputColumnNames) == 0 or len(data) < 2:
        np_projection = default_projection()
    else:
        if inputCardinality == "single":
            path = _dict_utils.split_escaped_string(inputColumnNames[0])
            np_array_of_embeddings = np.stack(
                _awl_pick(data, path).to_numpy(False), axis=0
            )
        else:
            column_data = [
                _awl_pick(data, _dict_utils.split_escaped_string(c))
                for c in inputColumnNames
            ]
            # Filter out null columns
            column_data = [
                col for col in column_data if not isinstance(col, pa.NullArray)
            ]
            np_array_of_embeddings = np.array(column_data).T
        # remove 0-only columns
        np_array_of_embeddings = np_array_of_embeddings[
            :, ~(np_array_of_embeddings.sum(axis=0) == 0)
        ]
        # If the selected data is not a 2D array, or if it has less than 2 columns, then
        # we can't perform a 2D projection. In this case, we just return a 2D array of
        # zeros.
        if (
            len(np_array_of_embeddings.shape) != 2
            or np_array_of_embeddings.shape[1] < 2
        ):
            np_projection = default_projection()
        else:
            np_array_of_embeddings = np.nan_to_num(np_array_of_embeddings, copy=False)
            np_projection = projection_utils.perform_2D_projection_with_timeout(
                np_array_of_embeddings, projectionAlgorithm, algorithmOptions
            )
    x_column = np_projection[:, 0]
    y_column = np_projection[:, 1]

    projection_column = pa.StructArray.from_arrays(
        [pa.array(x_column), pa.array(y_column)], ["x", "y"]
    )

    projection_column = pa.StructArray.from_arrays(
        [projection_column, source_column], ["projection", "source"]
    )

    return ArrowWeaveList(
        projection_column,
        types.TypedDict(
            {
                "projection": types.TypedDict(
                    {
                        "x": types.Number(),
                        "y": types.Number(),
                    }
                ),
                "source": table.object_type,
            }
        ),
        table._artifact,
    )


@op(
    name="ArrowWeaveListTypedDict-keys",
    input_type={"self": ArrowWeaveListType(types.TypedDict({}))},
    output_type=lambda input_types: ArrowWeaveListType(types.List(types.String())),
)
def keys(self):
    keys = list(self.object_type.property_types.keys())
    return convert.to_arrow(
        [keys] * len(self), types.List(types.List(types.String())), self._artifact
    )


@op(
    name="ArrowWeaveListTypedDict-columnNames",
    hidden=True,
    input_type={"self": ArrowWeaveListType(types.TypedDict({}))},
    output_type=lambda input_types: ArrowWeaveListType(types.String()),
)
def columnNames(self):
    return convert.to_arrow(
        list(self.object_type.property_types.keys()),
        types.List(types.String()),
        self._artifact,
    )


def _with_column_type(cur_type: types.Type, key: str, new_col_type: types.Type):
    if not isinstance(cur_type, types.TypedDict):
        cur_type = types.TypedDict()
    path = _dict_utils.split_escaped_string(key)

    property_types = {**cur_type.property_types}

    if len(path) > 1:
        return _with_column_type(
            cur_type,
            path[0],
            _with_column_type(
                property_types[path[0]], ".".join(path[1:]), new_col_type
            ),
        )

    col_names = list(property_types.keys())
    try:
        key_index = col_names.index(key)
    except ValueError:
        key_index = None
    if key_index is not None:
        property_types.pop(key)
        col_names.pop(key_index)
    property_types[key] = new_col_type
    return types.TypedDict(property_types)


def _with_columns_output_type(input_type):
    new_col_types = input_type["cols"]
    obj_type = input_type["self"].object_type
    for k, v in new_col_types.property_types.items():
        # added column type is optional, because we may need to extend with None
        # if it is too short.
        obj_type = _with_column_type(obj_type, k, v.object_type)
    return ArrowWeaveListType(obj_type)


@op(
    name="ArrowWeaveListTypedDict-with_columns",
    hidden=True,
    input_type={
        "self": ArrowWeaveListType(types.TypedDict({})),
        "cols": types.Dict(types.String(), ArrowWeaveListType()),
    },
    output_type=_with_columns_output_type,
)
def with_columns(self, cols):
    for col_val in cols.values():
        if len(self) != len(col_val):
            return None
    return self.with_columns(cols)



import typing

from . import context_state
from . import ref_util


def make_object_getattribute(
    allowed_attributes: list[str],
) -> typing.Callable[[typing.Any, str], typing.Any]:
    # Weave objects must auto-dereference refs when they are accessed.
    def object_getattribute(self: typing.Any, name: str) -> typing.Any:
        attribute = object.__getattribute__(self, name)
        if name not in allowed_attributes:
            return attribute
        return ref_util.val_with_relative_ref(
            self, attribute, [ref_util.OBJECT_ATTRIBUTE_EDGE_TYPE, str(name)]
        )

    return object_getattribute


def make_object_lookup_path() -> typing.Callable[
    [typing.Any, typing.List[str]], typing.Any
]:
    def object_lookup_path(self: typing.Any, path: typing.List[str]) -> typing.Any:
        assert len(path) > 1
        edge_type = path[0]
        edge_path = path[1]
        assert edge_type == ref_util.OBJECT_ATTRIBUTE_EDGE_TYPE

        res = getattr(self, edge_path)
        remaining_path = path[2:]
        if remaining_path:
            return res._lookup_path(remaining_path)
        return res

    return object_lookup_path


def build_ref_aware_object_subclass(
    target_name: str, starting_class: type, allowed_attributes: list[str]
) -> type:
    return type(
        target_name,
        (starting_class,),
        {
            "__getattribute__": make_object_getattribute(allowed_attributes),
            "_lookup_path": make_object_lookup_path(),
        },
    )



import dataclasses
import typing
import weave

from .. import panel
from .. import graph

from .panel_group import PanelBankSectionConfig

RenderType = typing.TypeVar("RenderType")


@weave.type()
class EachColumnConfig(typing.Generic[RenderType]):
    layoutMode: str = dataclasses.field(default_factory=lambda: "flow")
    pbLayoutConfig: typing.Optional[PanelBankSectionConfig] = dataclasses.field(
        default_factory=lambda: None
    )
    render: RenderType = dataclasses.field(default_factory=lambda: graph.VoidNode())  # type: ignore


@weave.type()
class EachColumn(panel.Panel):
    id = "EachColumn"
    config: typing.Optional[EachColumnConfig] = dataclasses.field(
        default_factory=lambda: None
    )



from pydantic import BaseModel, create_model
from . import weave_types as types
import typing
import enum


class JSONType(enum.Enum):
    STRING = "string"
    NUMBER = "number"
    INTEGER = "integer"
    BOOLEAN = "boolean"
    OBJECT = "object"
    ARRAY = "array"
    NULL = "null"


class JSONSchema(typing.TypedDict, total=False):
    title: str
    type: JSONType
    properties: "typing.Optional[typing.Dict[str, JSONSchema]]"
    required: typing.Optional[typing.List[str]]
    items: "typing.Optional[typing.Union[JSONSchema, JSONType]]"  # Type of array items


def weave_type_to_pydantic(
    property_types: dict[str, types.Type], name: str
) -> BaseModel:
    field_types = {}
    for k, v in property_types.items():
        if isinstance(v, types.TypedDict):
            field_types[k] = weave_type_to_pydantic(v.property_types, "TypedDict")
        else:
            instance_classes = v._instance_classes()
            if not instance_classes:
                raise ValueError(f"Cannot convert {v} to pydantic")
            field_types[k] = instance_classes[0]  # type: ignore
    field_defs = {k: (v, ...) for k, v in field_types.items()}
    return create_model(name, **field_defs)  # type: ignore


def json_schema_to_weave_type(json_schema: JSONSchema) -> types.Type:
    type = JSONType(json_schema["type"])
    if type == JSONType.INTEGER:
        return types.Int()
    elif type == JSONType.NUMBER:
        return types.Number()
    elif type == JSONType.STRING:
        return types.String()
    elif type == JSONType.NULL:
        return types.NoneType()
    elif type == JSONType.BOOLEAN:
        return types.Boolean()
    elif type == JSONType.ARRAY:
        items_type = json_schema["items"]
        assert items_type is not None, "No items type given for array"
        if not isinstance(items_type, dict):
            items_type = {"type": items_type}
        return types.List(json_schema_to_weave_type(items_type))
    elif type == JSONType.OBJECT:

        properties = json_schema.get("properties", {}) or {}
        weave_properties = {
            key: json_schema_to_weave_type(value) for key, value in properties.items()
        }

        required_keys: set[str] = set(json_schema.get("required", []) or [])
        not_required = set(properties.keys()) - required_keys

        return types.TypedDict(weave_properties, not_required_keys=not_required)
    raise ValueError(f"Could not get weave type for json schema {json_schema}")



import inspect
import typing
import dataclasses


from . import weave_types as types
from . import infer_types
from . import decorator_class
from . import errors
from . import context_state
from . import object_type_ref_util

_py_type = type


def type(
    __override_name: typing.Optional[str] = None,
    __is_simple: bool = False,
    __init: typing.Optional[bool] = None,
    __mixins: typing.Optional[list[typing.Type]] = None,
):
    def wrap(target):
        init = False
        if __init is not None:
            init = __init
        else:
            init = getattr(target.__bases__[0], "__init", True)
        target.__init = init
        dc = dataclasses.dataclass(target, init=init)
        fields = dataclasses.fields(dc)
        target_name = target.__name__
        if __override_name is not None:
            target_name = __override_name
        dc = object_type_ref_util.build_ref_aware_object_subclass(
            target.__name__,
            dc,
            [field.name for field in fields],
        )
        base_type = types.ObjectType
        if target.__bases__:
            # Add the first base classes as the type base.
            # TODO: should we add all bases?
            target_base0 = target.__bases__[0]
            if hasattr(target_base0, "WeaveType"):
                base_type = target_base0.WeaveType

        if __is_simple:
            bases = (
                base_type,
                types._PlainStringNamedType,
            )
        else:
            bases = (base_type,)

        if __mixins is not None:
            bases = tuple(__mixins) + bases

        TargetType = _py_type(f"{target_name}Type", bases, {})
        TargetType.name = target_name
        TargetType.instance_classes = dc
        TargetType.instance_class = dc

        type_vars: dict[str, types.Type] = {}
        static_property_types: dict[str, types.Type] = {}
        for field in fields:
            if isinstance(field.type, typing.TypeVar):
                # This is a Python type variable
                type_vars[field.name] = types.Any()
            else:
                try:
                    weave_type = infer_types.python_type_to_type(field.type)
                except TypeError:
                    # hmmm... Exception rewriting. Am I OK with this? Could be overly aggressive.
                    # TODO: decide if we should do this
                    raise errors.WeaveDefinitionError(
                        f"{target}.{field.name} is not a valid python type (a class or type)"
                    )
                # if weave_type == types.UnknownType():
                #     raise errors.WeaveDefinitionError(
                #         f"Weave doesn't yet handle the type '{field.type}' at {target}.{field.name}"
                #     )

                if types.type_is_variable(weave_type):
                    # this is a Weave type with a type variable in it
                    type_vars[field.name] = weave_type
                else:
                    static_property_types[field.name] = weave_type

        relocatable = not context_state.get_loading_built_ins()

        # Iterate through methods, finding ops that have versions (not builtins)
        # and sticking them on the type. This way we'll serialize
        # and deserialize them along with the data attached to the object
        if relocatable:
            for name, member in inspect.getmembers(target):
                from . import op_def
                from . import op_def_type

                if isinstance(member, op_def.BoundOpDef):
                    static_property_types[name] = op_def_type.OpDefType()

        if type_vars:
            setattr(TargetType, "__annotations__", {})
        for name, default_type in type_vars.items():
            setattr(TargetType, name, default_type)
            TargetType.__dict__["__annotations__"][name] = types.Type

        def property_types_method(self):
            property_types = {}
            for name in type_vars:
                property_types[name] = getattr(self, name)
            for name, prop_type in static_property_types.items():
                property_types[name] = prop_type
            return property_types

        TargetType.property_types = property_types_method

        if not type_vars:
            # workaround rich bug. dataclasses __repr__ doesn't print if the dataclass has no
            # fields. Insert our own __repr__ in that case so rich doesn't try to do its
            # special dataclass repr.
            TargetType.__repr__ = lambda self: f"{self.__class__.__name__}()"

        TargetType = dataclasses.dataclass(frozen=True)(TargetType)

        TargetType._relocatable = relocatable

        dc.WeaveType = TargetType
        decorator_class.weave_class(weave_type=TargetType)(dc)

        # constructor op for this type. due to a circular dependency with ArrowWeave* types, we
        # define the vectorized constructor ops in vectorize.py instead of here

        # Generating this constructor is currently disabled, now that we have reloctable object
        # types and we serialize methods as well, we don't want to serialize the constructor.
        # Instead we could have a generic constructor for all objects, but it's unclear how this
        # constructor will need to be used so for now we go with nothing.

        # @decorator_op.op(
        #     name=f"objectConstructor-_new_{target_name.replace('-', '_')}",
        #     input_type={
        #         field.name: static_property_types.get(field.name, None)
        #         or type_vars[field.name]
        #         for field in fields
        #     },
        #     output_type=TargetType(),
        #     render_info={"type": "function"},
        # )
        # def constructor(**attributes):
        #     return dc(
        #         **{field.name: attributes[field.name] for field in fields if field.init}
        #     )

        # dc.constructor = constructor

        return dc

    return wrap
